import pytest


def retrieval_baseline_training():
    from pytorch_lightning import Trainer
    from massspecgym.data import RetrievalDataset, MassSpecDataModule
    from massspecgym.data.transforms import SpecTokenizer, MolFingerprinter
    from massspecgym.models.retrieval import DeepSetsRetrieval

    # Load dataset
    dataset = RetrievalDataset(
        spec_transform=SpecTokenizer(n_peaks=60),
        mol_transform=MolFingerprinter(),
    )

    # Init data module
    data_module = MassSpecDataModule(
        dataset=dataset,
        batch_size=3
    )

    # Init model
    model = DeepSetsRetrieval(
        bootstrap_metrics=True,
        out_channels=2048,
        fourier_features=True
    )

    # Init trainer
    trainer = Trainer(
        accelerator="cpu", max_epochs=1, log_every_n_steps=1,
        limit_train_batches=3, limit_val_batches=3
    )

    trainer.fit(model, datamodule=data_module)


def test_retrieval_baseline_training():
    try:
        retrieval_baseline_training()
    except Exception as e:
        pytest.fail(f"retrieval_baseline_training() raised an exception: {e}")



import copy
import math
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch_geometric as pyg

from massspecgym.simulation_utils.misc_utils import safelog


def get_clones(module, N):
	return nn.ModuleList([copy.deepcopy(module) for i in range(N)])

def build_lr_scheduler(
	optimizer, decay_rate: float, decay_steps: int = 5000, warmup_steps: int = 1000
):
	"""build_lr_scheduler.

	Args:
		optimizer:
		decay_rate (float): decay_rate
		decay_steps (int): decay_steps
		warmup_steps (int): warmup_steps
	"""

	def lr_lambda(step):
		if step >= warmup_steps:
			# Adjust
			step = step - warmup_steps
			rate = decay_rate ** (step // decay_steps)
		else:
			rate = 1 - math.exp(-step / warmup_steps)
		return rate

	scheduler = torch.optim.lr_scheduler.LambdaLR(optimizer, lr_lambda=lr_lambda)
	return scheduler


class ZerosAggregation(nn.Module):
	""" Zero Aggreation module
		Just sums up features acroess all elements
	Args:
		nn (_type_): _description_
	"""
	def __init__(self):
		super().__init__()
		self.dummy_agg = pyg.nn.SumAggregation()

	def forward(self, x, batch):
		# sum x for the indices of elements for applying the aggregation. 
		# indices here work the same as reduce/sum
		agg = self.dummy_agg(x, batch)
		return 0.*agg


def build_pool_module(pool_type: str, node_dim: int, num_step_set2set:int=3):
	"""medthos build pooling layer for given pool_type and dim
		node dim is required for attention layer. 
		if pool type is not one of 'sum','mean','max','attention', return ZerosAggregation
	Args:
		pool_type (str): pooling type in ['sum','mean','max','attention']
		node_dim (int): node_dim, used for attention

	Returns:
		_type_: _description_
	"""

	if pool_type == "sum":
		pool_module = pyg.nn.SumAggregation()
	elif pool_type == "mean":
		pool_module = pyg.nn.MeanAggregation()
	elif pool_type == "max":
		pool_module = pyg.nn.MaxAggregation()
	elif pool_type == "attention":
		pool_module = pyg.nn.AttentionalAggregation(
			gate_nn = nn.Sequential(
				nn.Linear(node_dim, node_dim),
				nn.ReLU(),
				nn.Linear(node_dim, 1)
			)
		)
	elif pool_type == "set2set":
		"With Set2Set for graph level prediction  (Order Matters: Sequence to sequence for sets) https://arxiv.org/abs/1511.06391"
		pool_module = nn.Sequential(
			pyg.nn.Set2Set(node_dim,processing_steps=num_step_set2set),
			nn.Linear(2 * node_dim, node_dim),
		)
	elif pool_type == "softmax":
		pool_module = pyg.nn.SoftmaxAggregation(learn=True)
	elif pool_type == "powermean":
		pool_module = pyg.nn.PowerMeanAggregation(learn=True)
	elif pool_type == "mean_std_softmax":
		raise NotImplementedError
		# TODO: this need hook up in the way that size of input need change respect this
		#pool_module = pyg.nn.MultiAggregation(['mean', 'std',  pyg.nn.SoftmaxAggregation(learn=True)])
	else:
		assert pool_type == "none", pool_type
		# just return 0s
		pool_module = ZerosAggregation()
	return pool_module

class GNN(nn.Module):
	""" Generic GNN Class
	"""

	def __init__(
		self,
		hidden_size: int,
		num_layers: int,
		node_feats_size: int,
		edge_feats_size: int,
		gnn_type: str,
		dropout: float,
		normalization: str,
		**kwargs
	):
		"""_summary_

		Args:
			hidden_size (int): _description_
			num_layers (int): _description_
			node_feats_size (int): _description_
			edge_feats_size (int): _description_
			gnn_type (str): _description_
			dropout (float): _description_

		Raises:
			NotImplementedError: _description_
		"""
		super().__init__()
		self.hidden_size = hidden_size
		# if include ce here, add ce value to each node
		self.edge_feats_size = edge_feats_size
		self.node_feats_size = node_feats_size
		self.dropout = dropout
		self.normalization = normalization

		self.gnn_type = gnn_type
		self.hidden_size = hidden_size
		self.num_layers = num_layers
		self.input_project = nn.Linear(self.node_feats_size, self.hidden_size)

		if self.gnn_type == "GINE":
			self.gnn = GINE(
				hidden_size=self.hidden_size,
				edge_feats_size=self.edge_feats_size,
				node_feats_size=self.node_feats_size,
				num_layers=self.num_layers,
				dropout=self.dropout,
				normalization=self.normalization
			)
		elif self.gnn_type == "NodeMLP":
			self.gnn = NodeMLP(
				hidden_size=self.hidden_size,
				edge_feats_size=self.edge_feats_size,
				node_feats_size=self.node_feats_size,
				num_layers=self.num_layers,
				dropout=self.dropout,
				normalization=self.normalization
			)
		elif self.gnn_type == "MPNN":
			self.gnn = MPNN(
				node_feats_size=self.node_feats_size,
				edge_feats_size=self.edge_feats_size,
				hidden_size=self.hidden_size,
				num_layers=self.num_layers,
				dropout=self.dropout,
				normalization=self.normalization
			)
		elif self.gnn_type == "GAT" or self.gnn_type == "GATv2":
			self.gnn = GAT(
				node_feats_size=self.node_feats_size,
				edge_feats_size=self.edge_feats_size,
				hidden_size=self.hidden_size,
				num_layers=self.num_layers,
				dropout=self.dropout,
				normalization=self.normalization,
				is_v2=(self.gnn_type == "GATv2")
			)
		else:
			raise NotImplementedError(self.gnn_type)

	def forward(self, x, batch_index, edge_index, edge_attr):
		"""encode batch of molecule graph"""
		z = self.input_project(x)
		z = self.gnn(z, batch_index, edge_index, edge_attr)
		return z

def get_norm(normalization,hidden_size):

	# norm_fn may or may not depend on the batch index
	if normalization == "batch":
		norm_mod = nn.BatchNorm1d(hidden_size)
		norm_fn = lambda x, b, n: n(x)
	elif normalization == "layer":
		norm_mod = nn.LayerNorm(hidden_size)
		norm_fn = lambda x, b, n: n(x)
	elif normalization == "graph":
		norm_mod = pyg.nn.GraphNorm(hidden_size)
		norm_fn = lambda x, b, n: n(x,b)
	else:
		assert normalization == "none", normalization
		norm_mod = nn.Identity()
		norm_fn = lambda x, b, n: n(x)	
	return norm_mod, norm_fn

class GINE(nn.Module):
	def __init__(
		self,
		hidden_size: int,
		node_feats_size: int,
		edge_feats_size: int,
		num_layers: int,
		dropout: float,
		normalization: str,
		**kwargs
	):
		"""GINE.
			Strategies for Pre-training Graph Neural Networks 
			https://arxiv.org/abs/1905.12265
		Args:
			hidden_size (int): Hidden layer size
			edge_feats (int): Number of edge feats. Must be onehot!
			num_layers (int): Number of message passing steps
			normalization (str): normalization type (batch, layer, none)
			dropout
		"""
		super().__init__()

		assert edge_feats_size >= 0, edge_feats_size
		self.edge_transform = nn.Linear(edge_feats_size, hidden_size)

		self.layers = []
		for i in range(num_layers):
			apply_fn = nn.Sequential(
				nn.Linear(hidden_size, hidden_size),
				nn.ReLU(),
				nn.Linear(hidden_size, hidden_size),
			)
			temp_layer = pyg.nn.conv.GINEConv(
				nn=apply_fn,
				eps=0.,
				edge_dim=None
			)
			self.layers.append(temp_layer)

		self.layers = nn.ModuleList(self.layers)
		# setup norms and dropout
		norm_mod, norm_fn = get_norm(normalization, hidden_size)
		self.norm_fn = norm_fn
		self.norms = get_clones(norm_mod, num_layers)
		self.dropouts = get_clones(nn.Dropout(dropout), num_layers)

	def forward(self, x, batch_index, edge_index, edge_attr):
		"""forward.
		"""
		edge_attr = self.edge_transform(edge_attr)
		for dropout, layer, norm in zip(self.dropouts, self.layers, self.norms):
			layer_out = layer(x, edge_index, edge_attr)
			layer_out = self.norm_fn(layer_out, batch_index, norm)
			x = F.relu(dropout(layer_out)) + x
		return x

class NodeMLP(nn.Module):

	def __init__(
		self,
		hidden_size: int,
		node_feats_size: int,
		edge_feats_size: int,
		num_layers: int,
		dropout: float,
		normalization: str,
		**kwargs
	):
		"""NodeMLP
		"""
		super().__init__()

		assert edge_feats_size == 0, edge_feats_size

		self.layers = []
		for i in range(num_layers):
			apply_fn = nn.Sequential(
				nn.Linear(hidden_size, hidden_size),
				nn.ReLU(),
				nn.Linear(hidden_size, hidden_size),
			)
			temp_layer = apply_fn
			self.layers.append(temp_layer)

		self.layers = nn.ModuleList(self.layers)
		# setup norm and norm_fn
		# norm_fn may or may not depend on the batch index
		if normalization == "batch":
			norm_mod = nn.BatchNorm1d(hidden_size)
			self.norm_fn = lambda x, b, n: n(x)
		elif normalization == "layer":
			norm_mod = nn.LayerNorm(hidden_size)
			self.norm_fn = lambda x, b, n: n(x)
		elif normalization == "graph":
			norm_mod = pyg.nn.GraphNorm(hidden_size)
			self.norm_fn = lambda x, b, n: n(x,b)
		else:
			assert normalization == "none", normalization
			norm_mod = nn.Identity()
			self.norm_fn = lambda x, b, n: n(x)
		self.norms = get_clones(norm_mod, num_layers)
		self.dropouts = get_clones(nn.Dropout(dropout), num_layers)

	def forward(self, x, batch_index, edge_index, edge_attr):
		"""forward.
		"""
		for dropout, layer, norm in zip(self.dropouts, self.layers, self.norms):
			layer_out = layer(x)
			layer_out = self.norm_fn(layer_out, batch_index, norm)
			x = F.relu(dropout(layer_out)) + x
		return x

class MPNN(nn.Module):
	def __init__(
		self,
		hidden_size: int,
		node_feats_size: int,
		edge_feats_size: int,
		num_layers: int,
		dropout: float,
		normalization: str,
		**kwargs
	):
		"""Neural Message Passing for Quantum Chemistry
			https://arxiv.org/abs/1704.01212
		"""
		super().__init__()
		assert edge_feats_size >= 0, edge_feats_size
		self.edge_transform = nn.Linear(edge_feats_size, hidden_size)
		# MPNN 
		self.layers = []
		for i in range(num_layers):
			edge_network = edge_network = nn.Sequential(
				nn.Linear(hidden_size, hidden_size), nn.ReLU(),
				nn.Linear(hidden_size, hidden_size * hidden_size))
			layer = pyg.nn.NNConv(
				hidden_size,
				hidden_size,
				edge_network,
				aggr='mean',
				root_weight=False)
			self.layers.append(layer)
		self.layers = nn.ModuleList(self.layers)
		# setup norms and dropout
		norm_mod, norm_fn = get_norm(normalization, hidden_size)
		self.norm_fn = norm_fn
		self.norms = get_clones(norm_mod, num_layers)
		self.dropouts = get_clones(nn.Dropout(dropout), num_layers)

	def forward(self, x, batch_index, edge_index, edge_attr):
		"""forward.
		"""
		edge_attr = self.edge_transform(edge_attr)
		for dropout, layer, norm in zip(self.dropouts, self.layers, self.norms):
			layer_out = layer(x, edge_index, edge_attr)
			layer_out = self.norm_fn(layer_out, batch_index, norm)
			x = F.relu(dropout(layer_out)) + x
		return x

class GAT(nn.Module):
	""" Graph attention networks. https://arxiv.org/abs/1710.10903 
		Graph attention networks v2. (How Attentive are Graph Attention Networks?  https://arxiv.org/abs/2105.14491)
	"""

	def __init__(
		self,
		node_feats_size: int,
		edge_feats_size: int,
		hidden_size: int,
		num_layers: int,
		dropout: float,
		normalization: str,
		num_gat_heads: int = 8,
		gat_dropout: float = 0.0,
		is_v2: bool = True,
		is_concat: bool = True,
		**kwargs
	):
		"""_summary_
		Args:
			node_feats_size (int): _description_
			edge_feats_size (int): _description_
			node_hidden_dim (int): _description_
			num_step_message_passing (int): _description_
			num_step_set2set (int): _description_
			output_dim (int): _description_
			gat_heads (int, optional): _description_. Defaults to 1.
			gat_dropout (float, optional): _description_. Defaults to 0.0.
			is_v2 (bool, optional). use v2 GATConv, Defaults to True
			is_concat (bool, optional). concatenated multihead attetion, Defaults to True
		"""
		super(GAT, self).__init__()
		assert edge_feats_size >= 0, edge_feats_size
		self.edge_transform = nn.Linear(edge_feats_size, hidden_size)
		gat_output_size = hidden_size // num_gat_heads if is_concat else hidden_size
		if is_concat and hidden_size % num_gat_heads != 0:
			raise ValueError(f"Ensure that the number of output channels of "
                             f"'GAT' (got '{hidden_size}') is divisible "
                             f"by the number of heads (got '{num_gat_heads}')")	
		GATConv_fn = pyg.nn.conv.GATConv if not is_v2 else pyg.nn.conv.GATv2Conv
		self.layers = []
		for i in range(num_layers):
			layer = GATConv_fn(
				in_channels=hidden_size,
				out_channels=gat_output_size,
				heads=num_gat_heads,
				edge_dim=hidden_size,
				dropout=gat_dropout,
				concat=is_concat)
			self.layers.append(layer)
		self.layers = nn.ModuleList(self.layers)
		# setup norms and dropout
		norm_mod, norm_fn = get_norm(normalization, hidden_size)
		self.norm_fn = norm_fn
		self.norms = get_clones(norm_mod, num_layers)
		self.dropouts = get_clones(nn.Dropout(dropout), num_layers)

	def forward(self, x, batch_index, edge_index, edge_attr):
		edge_attr = self.edge_transform(edge_attr)
		for dropout, layer, norm in zip(self.dropouts, self.layers, self.norms):
			layer_out = layer(x, edge_index, edge_attr)
			layer_out = self.norm_fn(layer_out, batch_index, norm)
			x = F.relu(dropout(layer_out)) + x
		return x
	
class MLPBlocks(nn.Module):
	"""Just Good Old Multilayer perceptron with residuals
		layer->dropout->activation(Relu)
		if residuals is True, add a skip connection between block
	Args:
		nn (_type_): _description_
	"""
	def __init__(
		self,
		input_size: int,
		hidden_size: int,
		normalization: str,
		dropout: float,
		num_layers: int,
		output_size: int = None,
		use_residuals: bool = False,
	):
		"""_summary_

		Args:
			input_size (int): input dimensions
			hidden_size (int): hidden_size layer dimensions
			dropout (float): dropout ratio
			num_layers (int): num of hidden layers + 1
			output_size (int, optional): output dimensions. Defaults to None.
			use_residuals (bool, optional): if residuals is True, add a skip connection between locks. Defaults to False.
		"""
		super().__init__()
		self.activation = nn.ReLU()
		self.dropout_layer = nn.Dropout(p=dropout)
		self.input_layer = nn.Linear(input_size, hidden_size)
		middle_layer = nn.Linear(hidden_size, hidden_size)
		self.layers = get_clones(middle_layer, num_layers - 1)

		self.output_layer = None
		self.output_size = output_size
		if self.output_size is not None:
			self.output_layer = nn.Linear(hidden_size, self.output_size)

		self.use_residuals = use_residuals

		# setup norm
		if normalization == "batch":
			norm_mod = nn.BatchNorm1d(hidden_size)
		elif normalization == "layer":
			norm_mod = nn.LayerNorm(hidden_size)
		else:
			assert normalization == "none", normalization
			norm_mod = nn.Identity()
		self.norms = get_clones(norm_mod, num_layers)

	def forward(self, x):
		output = x
		output = self.input_layer(x)
		output = self.dropout_layer(output)
		output = self.activation(output)
		old_op = output
		for layer_index, layer in enumerate(self.layers):
			output = layer(output)
			output = self.dropout_layer(output)
			output = self.activation(output)
			output = self.norms[layer_index](output)
			if self.use_residuals:
				output = output + old_op
				old_op = output

		if self.output_layer is not None:
			output = self.output_layer(output)
		return output


class NeimsBlock(nn.Module):
	""" from the NEIMS paper (uses LeakyReLU instead of ReLU) """

	def __init__(self, in_dim, out_dim, dropout):

		super().__init__()
		bottleneck_factor = 0.5
		bottleneck_size = int(round(bottleneck_factor * out_dim))
		self.in_batch_norm = nn.BatchNorm1d(in_dim)
		self.in_activation = nn.LeakyReLU()
		self.in_linear = nn.Linear(in_dim, bottleneck_size)
		self.out_batch_norm = nn.BatchNorm1d(bottleneck_size)
		self.out_linear = nn.Linear(bottleneck_size, out_dim)
		self.out_activation = nn.LeakyReLU()
		self.dropout = nn.Dropout(p=dropout)

	def forward(self, x):

		h = x
		h = self.in_batch_norm(h)
		h = self.in_activation(h)
		h = self.dropout(h)
		h = self.in_linear(h)
		h = self.out_batch_norm(h)
		h = self.out_activation(h)
		h = self.out_linear(h)
		return h


class SpecFFN(nn.Module):

	def __init__(self, 
		input_size, 
		hidden_size,
		mz_max,
		mz_bin_res,
		num_layers,
		dropout, 
		prec_mz_offset,
		bidirectional,
		use_residuals,
		output_map_size):

		super().__init__()

		self.input_size = input_size
		self.mz_max = mz_max
		self.mz_bin_res = mz_bin_res
		self.prec_mz_offset = prec_mz_offset
		self.bidirectional = bidirectional
		self.use_residuals = use_residuals
		self._compute_output_size()

		self.in_layer = nn.Linear(input_size, hidden_size)
		self.ff_layers = nn.ModuleList([])
		# self.ff_layers.append(nn.Linear(mlp_hidden_size, mlp_hidden_size))
		for i in range(num_layers):
			self.ff_layers.append(
				NeimsBlock(
					hidden_size,
					hidden_size,
					dropout))
		if output_map_size == -1:
			if self.bidirectional:
				# assumes gating, mass masking
				self.forw_out_layer = nn.Linear(hidden_size, self.output_size)
				self.rev_out_layer = nn.Linear(hidden_size, self.output_size)
				self.out_gate = nn.Sequential(
					*[nn.Linear(hidden_size, self.output_size), nn.Sigmoid()])
			else:
				self.out_layer = nn.Linear(hidden_size, self.output_size)
				self.out_gate = nn.Sequential(
					*[nn.Linear(hidden_size, self.output_size), nn.Sigmoid()])
		else:
			if self.bidirectional:
				# assumes gating, mass masking
				self.forw_out_layer = LowRankDense(hidden_size, self.output_size, output_map_size)
				self.rev_out_layer = LowRankDense(hidden_size, self.output_size, output_map_size)
				self.out_gate = nn.Sequential(
					*[LowRankDense(hidden_size, self.output_size, output_map_size), nn.Sigmoid()])
			else:
				self.out_layer = LowRankDense(hidden_size, self.output_size, output_map_size)
				self.out_gate = nn.Sequential(
					*[LowRankDense(hidden_size, self.output_size, output_map_size), nn.Sigmoid()])
		self.out_activation = nn.ReLU()
		self.out_normalization = lambda x: F.normalize(x, p=1, dim=1)

	def _compute_output_size(self):

		self.mz_bins = nn.Parameter(torch.arange(self.mz_bin_res,self.mz_max+self.mz_bin_res,self.mz_bin_res))
		self.mzs = nn.Parameter(self.mz_bins-0.5*self.mz_bin_res)
		self.output_size = self.mzs.shape[0]

	def _prec_mz_to_idx(self,prec_mz):

		prec_mz_idx = torch.bucketize(torch.clamp(prec_mz,min=0.,max=self.mz_max-self.mz_bin_res),self.mz_bins.to(prec_mz.device),right=True)
		assert torch.max(prec_mz_idx) < self.output_size, (prec_mz_idx,self.output_size)
		return prec_mz_idx

	def forward(self, input_h, prec_mz):

		# get prec_mz_idxs
		prec_mz_idxs = self._prec_mz_to_idx(prec_mz)
		# process inputs
		fh = self.in_layer(input_h)
		# big MLP
		for layer in self.ff_layers:
			if self.use_residuals:
				fh = fh + layer(fh)
			else:
				fh = layer(fh)
		# bidirectional prediction
		if self.bidirectional:
			ff = self.forw_out_layer(fh)
			fr = reverse_prediction(
				self.rev_out_layer(fh),
				prec_mz_idxs,
				self.prec_mz_offset)
			fg = self.out_gate(fh)
			fo = ff * fg + fr * (1. - fg)
			fo = mask_prediction_by_mass(
				fo, prec_mz_idxs, self.prec_mz_offset)
		else:
			# apply output layer
			fo = self.out_layer(fh)
			# apply gating
			fg = self.out_gate(fh)
			fo = fg * fo
		fo = self.out_activation(fo)
		spec = self.out_normalization(fo)
		# handle all zeroes (set to first bin by default)
		all_zero_mask = torch.max(spec,dim=1)[0]<=0.
		all_zero_bonus = torch.zeros_like(spec)
		all_zero_bonus[all_zero_mask,0] = 1.
		spec = spec + all_zero_bonus
		# convert dense spectrum to sparse
		mask = spec>0.
		pred_mzs = (self.mzs.unsqueeze(0).expand(spec.shape[0],-1))[mask]
		pred_logprobs = safelog(spec)[mask]
		pred_batch_idxs = torch.arange(spec.shape[0],device=spec.device).unsqueeze(1).expand(-1,spec.shape[1])[mask]
		# pred_specs = spec
		return pred_mzs, pred_logprobs, pred_batch_idxs #, pred_specs


def mask_prediction_by_mass(raw_prediction, prec_mass_idx, prec_mass_offset):
	# adapted from NEIMS
	# raw_prediction is [B,D], prec_mass_idx is [B]

	max_idx = raw_prediction.shape[1]
	assert torch.all(prec_mass_idx < max_idx)
	idx = torch.arange(max_idx, device=prec_mass_idx.device)
	mask = (
		idx.unsqueeze(0) <= (
			prec_mass_idx.unsqueeze(1) +
			prec_mass_offset)).float()
	return mask * raw_prediction


def reverse_prediction(raw_prediction, prec_mass_idx, prec_mass_offset):
	# adapted from NEIMS
	# raw_prediction is [B,D], prec_mass_idx is [B]

	batch_size = raw_prediction.shape[0]
	max_idx = raw_prediction.shape[1]
	assert torch.all(prec_mass_idx < max_idx)
	rev_prediction = torch.flip(raw_prediction, dims=(1,))
	# convention is to shift right, so we express as negative to go left
	offset_idx = torch.minimum(
		max_idx * torch.ones_like(prec_mass_idx),
		prec_mass_idx + prec_mass_offset + 1)
	shifts = - (max_idx - offset_idx)
	gather_idx = torch.arange(
		max_idx,
		device=raw_prediction.device).unsqueeze(0).expand(
		batch_size,
		max_idx)
	gather_idx = (gather_idx - shifts.unsqueeze(1)) % max_idx
	offset_rev_prediction = torch.gather(rev_prediction, 1, gather_idx)
	# you could mask_prediction_by_mass here but it's unnecessary
	return offset_rev_prediction


def nan_forward_hook(self, input, output):
	if isinstance(output, tuple):
		outputs = list(output)
	elif isinstance(output, dict):
		outputs = list(output.values())
	else:
		outputs = [output]
	for i, val in enumerate(outputs):
		nan_mask = torch.isnan(val)
		if nan_mask.any():
			print(">> In", self.__class__.__name__)
			raise RuntimeError(f"Found NAN in output {i} at indices: ", nan_mask.nonzero(), "where:", val[nan_mask.nonzero()[:, 0].unique(sorted=True)])


def nan_backward_hook(self, grad_input, grad_output):
	for i, val in enumerate(grad_input):
		if val is None:
			continue
		nan_mask = torch.isnan(val)
		if nan_mask.any():
			print(">> In", self.__class__.__name__)
			raise RuntimeError(f"Found NAN in grad_input {i} at indices: ", nan_mask.nonzero(), "where:", val[nan_mask.nonzero()[:, 0].unique(sorted=True)])
	for i, val in enumerate(grad_output):
		if val is None:
			continue
		nan_mask = torch.isnan(val)
		if nan_mask.any():
			print(">> In", self.__class__.__name__)
			raise RuntimeError(f"Found NAN in grad_output {i} at indices: ", nan_mask.nonzero(), "where:", val[nan_mask.nonzero()[:, 0].unique(sorted=True)])


class LowRankDense(nn.Module):

    def __init__(self, input_dim, output_dim, rank):
        super(LowRankDense, self).__init__()
        self.layer1 = nn.Linear(input_dim, rank, bias=False)
        self.layer2 = nn.Linear(rank, output_dim)

    def forward(self, x):
        x = self.layer1(x)
        x = self.layer2(x)
        return x



import time
import unittest
from collections.abc import Generator

import torch
from rdkit import Chem
from rdkit.Chem import Draw
import pickle

from massspecgym.models.de_novo import RandomDeNovo
from massspecgym.models.de_novo.random import AtomWithValence, ValenceAndCharge


class RandomDeNovoTestcase(unittest.TestCase):
    def setUp(self, draw_molecules: bool = False) -> None:
        self.generator_with_formula = RandomDeNovo(
            formula_known=True, cache_results=False
        )
        self.generator_without_formula = RandomDeNovo(
            formula_known=True, cache_results=False
        )
        self.draw_molecules = draw_molecules
        with open("data/misc/random_de_novo_element_stats_from_trn_with_hydrogens.pkl", "rb") as file:
            self.element_2_bond_stats = pickle.load(file)


    def test_generator_of_element_atoms_split_into_valence_groups_1(self):
        self.assertTrue(
            isinstance(
                self.generator_with_formula.generator_for_splits_of_chem_element_atoms_by_possible_valences(
                    atom_type="N",
                    possible_valences=[3, 5],
                    atom_count=22,
                    already_assigned_groups_of_atoms=dict(),
                ),
                Generator,
            )
        )

    def test_generator_of_element_atoms_split_into_valence_groups_2(self):
        self.assertEqual(
            len(
                list(
                    self.generator_with_formula.generator_for_splits_of_chem_element_atoms_by_possible_valences(
                        atom_type="N",
                        possible_valences=[3, 5],
                        atom_count=22,
                        already_assigned_groups_of_atoms=dict(),
                    )
                )
            ),
            23,
        )

    def test_generator_of_element_atoms_split_into_valence_groups_3(self):
        self.assertEqual(
            len(
                list(
                    self.generator_with_formula.generator_for_splits_of_chem_element_atoms_by_possible_valences(
                        atom_type="Mn",
                        possible_valences=[7, 4, 2],
                        atom_count=22,
                        already_assigned_groups_of_atoms=dict(),
                    )
                )
            ),
            276,
        )

    def test_valence_assignment_1(self):
        valence_assignment_generator = (
            self.generator_with_formula.assigner_of_valences_to_all_atoms(
                unassigned_molecule_elements_with_counts={"C": 1, "O": 2},
                already_assigned_atoms_with_valences={},
                common_valences_only=True,
            )
        )
        all_common_valence_possibilities = list(valence_assignment_generator)
        self.assertEqual(3, len(all_common_valence_possibilities))

    def test_valence_assignment_2(self):
        valence_assignment_generator = (
            self.generator_with_formula.assigner_of_valences_to_all_atoms(
                unassigned_molecule_elements_with_counts={"C": 1, "O": 2},
                already_assigned_atoms_with_valences={},
                common_valences_only=True,
            )
        )
        valence_assignment = list(valence_assignment_generator)[0]
        self.assertEqual(len(valence_assignment), 2)

    def test_valence_assignment_3(self):
        valence_assignment_generator = (
            self.generator_with_formula.assigner_of_valences_to_all_atoms(
                unassigned_molecule_elements_with_counts={"C": 1, "O": 2},
                already_assigned_atoms_with_valences={},
                common_valences_only=True,
            )
        )
        valence_assignment = list(valence_assignment_generator)[0]
        atom_types = {atom.atom_type for atom in valence_assignment.keys()}
        self.assertEqual({"C", "O"}, atom_types)

    def test_valence_assignment_4(self):
        valence_assignment_generator = (
            self.generator_with_formula.assigner_of_valences_to_all_atoms(
                unassigned_molecule_elements_with_counts={
                    "C": 7,
                    "H": 5,
                    "N": 3,
                    "O": 6,
                },
                already_assigned_atoms_with_valences={},
                common_valences_only=True,
            )
        )
        self.assertEqual(28, len(list(valence_assignment_generator)))

    def test_valence_assignment_5(self):
        valence_assignment_generator = (
            self.generator_with_formula.assigner_of_valences_to_all_atoms(
                unassigned_molecule_elements_with_counts={"C": 17, "H": 14, "O": 4},
                already_assigned_atoms_with_valences={},
                common_valences_only=True,
            )
        )
        self.assertEqual(5, len(list(valence_assignment_generator)))

    def test_valence_feasibility_check_too_few_edges(self):
        valence_assignment = {
            AtomWithValence(
                atom_type="C", atom_valence_and_charge=ValenceAndCharge(1, 0)
            ): 7,
            AtomWithValence(
                atom_type="H", atom_valence_and_charge=ValenceAndCharge(1, 0)
            ): 5,
            AtomWithValence(
                atom_type="N", atom_valence_and_charge=ValenceAndCharge(1, 0)
            ): 3,
            AtomWithValence(
                atom_type="O", atom_valence_and_charge=ValenceAndCharge(1, 0)
            ): 6,
        }
        is_assignment_feasible = (
            self.generator_with_formula.is_valence_assignment_feasible(
                valence_assignment
            )
        )
        self.assertFalse(is_assignment_feasible)

    def test_valence_feasibility_check_odd_degrees_sum(self):
        valence_assignment = {
            AtomWithValence(
                atom_type="C", atom_valence_and_charge=ValenceAndCharge(3, -1)
            ): 1,
            AtomWithValence(
                atom_type="C", atom_valence_and_charge=ValenceAndCharge(4, 0)
            ): 1,
            AtomWithValence(
                atom_type="H", atom_valence_and_charge=ValenceAndCharge(1, 0)
            ): 6,
        }
        is_assignment_feasible = (
            self.generator_with_formula.is_valence_assignment_feasible(
                valence_assignment
            )
        )
        self.assertFalse(is_assignment_feasible)

    def test_sampling_of_atoms_by_valence_partition(self):
        self.assertEqual(
            len(
                self.generator_with_formula.get_feasible_atom_valence_assignments(
                    "C23H27N5O2"
                )
            ),
            3,
        )

    def test_random_molecule_generation_hard(self):
        start_time = time.time()
        molecule = (
            self.generator_with_formula.generate_random_molecule_graphs_via_traversal(
                chemical_formula="C13H24Cl6O8P2"
            )
        )
        total_secs = time.time() - start_time
        self.assertLess(total_secs, 1)

    def test_random_molecule_generation_1(self):
        for gen_i in range(50):
            molecules = self.generator_with_formula.generate_random_molecule_graphs_via_traversal(
                chemical_formula="C23H17Cl2N5O4"
            )
            if self.draw_molecules:
                for mol_i, molecule in enumerate(molecules):
                    img = Draw.MolToImage(molecule, size=(700, 700))
                    img.save(f"molecule_1_{gen_i}_{mol_i}.png")

    def test_random_molecule_generation_2(self):
        for gen_i in range(50):
            molecules = self.generator_with_formula.generate_random_molecule_graphs_via_traversal(
                chemical_formula="C20H36O7P2"
            )
            if self.draw_molecules:
                for mol_i, molecule in enumerate(molecules):
                    img = Draw.MolToImage(molecule, size=(700, 700))
                    img.save(f"molecule_2_{gen_i}_{mol_i}.png")

    def test_step_function(self):
        batch = {
            "mol": [
                "C/C1=C/CC[C@@]2(C)O[C@@H]2[C@H]2OC(=O)[C@H](CN(C)C)[C@@H]2CC1",
                "COc1ncc2cc(C(=O)Nc3c(Cl)ccc(C(=O)NCc4cc(Cl)ccc4)c3)c(=O)[nH]c2n1",
                "CNC(=O)O[C@H]1COc2c(cc(N3CCN(C4COC4)CC3)cc2)[C@@H]1NC(=O)c1ccc(F)cc1",
                "COc1nc(N2CCC3(CCCN(Cc4c[nH]c5ccccc45)C3=O)CC2)ncc1",
                "Cc1c(C)c2c(cc1)c(=O)c1cccc(CC(=O)O)c1o2",
            ]
        }
        for batch_i in range(50):
            mol_preds = self.generator_with_formula.step(batch)["mols_pred"]
            if self.draw_molecules:
                for input_mol_i, molecule_smiles in enumerate(mol_preds):
                    for mol_i, _smiles in enumerate(molecule_smiles):
                        molecule = Chem.MolFromSmiles(_smiles)
                        img = Draw.MolToImage(molecule, size=(700, 700))
                        img.save(f"step_molecule_{input_mol_i}_{mol_i}.png")

    def test_single_candidate_generation(self):
        generator = RandomDeNovo(formula_known=True, max_top_k=1)
        batch = {
            "mol": [
                "CCCC[C@@H](C)[C@H]([C@H](C[C@@H](C)C[C@@H](CCCCCC[C@@H]([C@@H](C)N)O)O)OC(=O)CC(CC(=O)O)C(=O)O)OC(=O)CC(CC(=O)O)C(=O)O",
            ]
        }
        for _ in range(100):
            self.assertEqual(1, len(generator.step(batch)["mols_pred"]))

    def test_slow_eval_molecule(self):
        molecules = (
            self.generator_with_formula.generate_random_molecule_graphs_via_traversal(
                chemical_formula="C23H17Cl2N5O4"
            )
        )

    def test_weight_recording(self):
        batch = {
            "mol": [
                "CC(C)C[C@H]1C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N1C)C(C)C)CC(C)C)C)C(C)C)CC(C)C)C)C(C)C)CC(C)C)C)C(C)C",
                "CC(C)C[C@H]1C(=O)N2CCC[C@H]2[C@]3(N1C(=O)[C@](O3)(C(C)C)NC(=O)[C@H]4CN([C@@H]5CC6=CNC7=CC=CC(=C67)C5=C4)C)O",
            ],
        }
        self.generator_without_formula.training_step(batch, batch_idx=torch.Tensor())
        self.assertEqual(
            {575.3107694120001: ["C32H41N5O5"], 908.6085741279999: ["C48H84N4O12"]},
            self.generator_without_formula.mol_weight_2_formulas,
        )

    def test_weight_recording_on_train_end(self):
        batch = {
            "mol": [
                "CC(C)C[C@H]1C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N1C)C(C)C)CC(C)C)C)C(C)C)CC(C)C)C)C(C)C)CC(C)C)C)C(C)C",
                "CC(C)C[C@H]1C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N1C)C(C)C)CC(C)C)C)C(C)C)CC(C)C)C)C(C)C)CC(C)C)C)C(C)C",
                "C",  # dummy for the test
                "CC(C)C[C@H]1C(=O)N2CCC[C@H]2[C@]3(N1C(=O)[C@](O3)(C(C)C)NC(=O)[C@H]4CN([C@@H]5CC6=CNC7=CC=CC(=C67)C5=C4)C)O",
            ],
        }
        self.generator_without_formula.training_step(batch, batch_idx=torch.Tensor())
        self.generator_without_formula.on_train_end()
        self.assertEqual(
            {
                16.031300127999998: [["CH4"], [1.0]],
                575.3107694120001: [["C32H41N5O5"], [1.0]],
                908.6085741279999: [["C48H84N4O12"], [1.0]],
            },
            self.generator_without_formula.mol_weight_2_formulas,
        )

    def test_sample_formula_with_the_closest_molecular_weight_1(self):
        batch = {
            "mol": [
                "CC(C)C[C@H]1C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N1C)C(C)C)CC(C)C)C)C(C)C)CC(C)C)C)C(C)C)CC(C)C)C)C(C)C",
                "CC(C)C[C@H]1C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N1C)C(C)C)CC(C)C)C)C(C)C)CC(C)C)C)C(C)C)CC(C)C)C)C(C)C",
                "C",  # dummy for the test
                "CC(C)C[C@H]1C(=O)N2CCC[C@H]2[C@]3(N1C(=O)[C@](O3)(C(C)C)NC(=O)[C@H]4CN([C@@H]5CC6=CNC7=CC=CC(=C67)C5=C4)C)O",
            ],
        }
        self.generator_without_formula.training_step(batch, batch_idx=torch.Tensor())
        self.generator_without_formula.on_train_end()

        self.assertEqual(
            "CH4",
            self.generator_without_formula.sample_formula_with_the_closest_molecular_weight(
                20
            ),
        )

    def test_sample_formula_with_the_closest_molecular_weight_2(self):
        batch = {
            "mol": [
                "CC(C)C[C@H]1C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N1C)C(C)C)CC(C)C)C)C(C)C)CC(C)C)C)C(C)C)CC(C)C)C)C(C)C",
                "CC(C)C[C@H]1C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N1C)C(C)C)CC(C)C)C)C(C)C)CC(C)C)C)C(C)C)CC(C)C)C)C(C)C",
                "C",  # dummy for the test
                "CC(C)C[C@H]1C(=O)N2CCC[C@H]2[C@]3(N1C(=O)[C@](O3)(C(C)C)NC(=O)[C@H]4CN([C@@H]5CC6=CNC7=CC=CC(=C67)C5=C4)C)O",
            ],
        }
        self.generator_without_formula.training_step(batch, batch_idx=torch.Tensor())
        self.generator_without_formula.on_train_end()

        self.assertEqual(
            "C32H41N5O5",
            self.generator_without_formula.sample_formula_with_the_closest_molecular_weight(
                300
            ),
        )

    def test_sample_formula_with_the_closest_molecular_weight_3(self):
        batch = {
            "mol": [
                "CC(C)C[C@H]1C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N1C)C(C)C)CC(C)C)C)C(C)C)CC(C)C)C)C(C)C)CC(C)C)C)C(C)C",
                "CC(C)C[C@H]1C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N1C)C(C)C)CC(C)C)C)C(C)C)CC(C)C)C)C(C)C)CC(C)C)C)C(C)C",
                "C",  # dummy for the test
                "CC(C)C[C@H]1C(=O)N2CCC[C@H]2[C@]3(N1C(=O)[C@](O3)(C(C)C)NC(=O)[C@H]4CN([C@@H]5CC6=CNC7=CC=CC(=C67)C5=C4)C)O",
            ],
        }
        self.generator_without_formula.training_step(batch, batch_idx=torch.Tensor())
        self.generator_without_formula.on_train_end()

        self.assertEqual(
            "C48H84N4O12",
            self.generator_without_formula.sample_formula_with_the_closest_molecular_weight(
                908.6085741279999
            ),
        )

    def test_chemical_elements_stats_computation_train_step(self):
        batch = {
            "mol": [
                "CC(C)C[C@H]1C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N1C)C(C)C)CC(C)C)C)C(C)C)CC(C)C)C)C(C)C)CC(C)C)C)C(C)C",
                "CC(C)C[C@H]1C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N1C)C(C)C)CC(C)C)C)C(C)C)CC(C)C)C)C(C)C)CC(C)C)C)C(C)C",
                "C",  # dummy for the test
                "CC(C)C[C@H]1C(=O)N2CCC[C@H]2[C@]3(N1C(=O)[C@](O3)(C(C)C)NC(=O)[C@H]4CN([C@@H]5CC6=CNC7=CC=CC(=C67)C5=C4)C)O",
            ],
        }
        generator_with_stats = RandomDeNovo(estimate_chem_element_stats=True)
        generator_with_stats.training_step(batch, batch_idx=torch.Tensor())

        self.assertEqual(
            118,
            generator_with_stats.element_2_bond_stats["C"][
                ValenceAndCharge(valence=4, charge=0)
            ][0][()][("C", 4, 0, 1.0)],
        )

    def test_chemical_elements_stats_computation_post_train(self):
        batch = {
            "mol": [
                "CC(C)C[C@H]1C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N1C)C(C)C)CC(C)C)C)C(C)C)CC(C)C)C)C(C)C)CC(C)C)C)C(C)C",
                "CC(C)C[C@H]1C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N([C@H](C(=O)O[C@@H](C(=O)N1C)C(C)C)CC(C)C)C)C(C)C)CC(C)C)C)C(C)C)CC(C)C)C)C(C)C",
                "C",  # dummy for the test
                "CC(C)C[C@H]1C(=O)N2CCC[C@H]2[C@]3(N1C(=O)[C@](O3)(C(C)C)NC(=O)[C@H]4CN([C@@H]5CC6=CNC7=CC=CC(=C67)C5=C4)C)O",
            ],
        }
        generator_with_stats = RandomDeNovo(estimate_chem_element_stats=True)
        generator_with_stats.training_step(batch, batch_idx=torch.Tensor())
        generator_with_stats.on_train_end()

        self.assertEqual(
            ([(1.0, 118), (2.0, 10)], 128),
            generator_with_stats.element_2_bond_stats["C"][
                ValenceAndCharge(valence=4, charge=0)
            ][0][()][
                AtomWithValence(
                    atom_type="C",
                    atom_valence_and_charge=ValenceAndCharge(valence=4, charge=0),
                )
            ],
        )

    def test_random_molecule_generation_with_stats_for_ion_1(self):
        generator_with_stats = RandomDeNovo(
            estimate_chem_element_stats=True,
            enforce_connectivity=False,
            cache_results=False,
        )
        generator_with_stats.element_2_bond_stats = self.element_2_bond_stats
        molecules = (
            self.generator_with_formula.generate_random_molecule_graphs_via_traversal(
                chemical_formula="C8H18OP+"
            )
        )

    def test_random_molecule_generation_with_stats_for_ion_2(self):
        generator_with_stats = RandomDeNovo(
            estimate_chem_element_stats=True,
            enforce_connectivity=False,
            cache_results=False,
        )
        generator_with_stats.element_2_bond_stats = self.element_2_bond_stats
        molecules = (
            self.generator_with_formula.generate_random_molecule_graphs_via_traversal(
                chemical_formula="C3H6NO4S-"
            )
        )

    def test_random_molecule_generation_with_stats_no_connectivity_1(self):
        generator_with_stats = RandomDeNovo(
            estimate_chem_element_stats=True,
            enforce_connectivity=False,
            cache_results=False,
        )
        generator_with_stats.element_2_bond_stats = self.element_2_bond_stats
        for gen_i in range(50):
            molecules = (
                generator_with_stats.generate_random_molecule_graphs_via_traversal(
                    chemical_formula="C23H17Cl2N5O4",
                )
            )
            if self.draw_molecules:
                for mol_i, molecule in enumerate(molecules):
                    img = Draw.MolToImage(molecule, size=(700, 700))
                    img.save(f"molecule_1_with_stats_{gen_i}_{mol_i}.png")

    def test_random_molecule_generation_with_stats_without_connectivity_2(self):
        generator_with_stats = RandomDeNovo(
            estimate_chem_element_stats=True,
            enforce_connectivity=False,
            cache_results=False,
        )
        generator_with_stats.element_2_bond_stats = self.element_2_bond_stats
        for gen_i in range(50):
            molecules = (
                generator_with_stats.generate_random_molecule_graphs_via_traversal(
                    chemical_formula="C20H36O7P2",
                )
            )
            if self.draw_molecules:
                for mol_i, molecule in enumerate(molecules):
                    img = Draw.MolToImage(molecule, size=(700, 700))
                    img.save(f"molecule_2_with_stats_{gen_i}_{mol_i}.png")


if __name__ == "__main__":
    unittest.main()



import pandas as pd
import json
import os
import typing as T
import numpy as np
import torch
import matchms
from pathlib import Path
from torch.utils.data.dataset import Dataset
from torch.utils.data.dataloader import default_collate
from torch_geometric.data import Data, Batch
from matchms.importing import load_from_mgf

import massspecgym.utils as utils
from massspecgym.data.transforms import SpecTransform, MolTransform, MolToInChIKey, MetaTransform
from massspecgym.simulation_utils.misc_utils import flatten_lol


class MassSpecDataset(Dataset):
    """
    Dataset containing mass spectra and their corresponding molecular structures. This class is
    responsible for loading the data from disk and applying transformation steps to the spectra and
    molecules.
    """

    def __init__(
        self,
        spec_transform: T.Optional[T.Union[SpecTransform, T.Dict[str, SpecTransform]]] = None,
        mol_transform: T.Optional[T.Union[MolTransform, T.Dict[str, MolTransform]]] = None,
        pth: T.Optional[Path] = None,
        return_mol_freq: bool = True,
        return_identifier: bool = True,
        identifiers_subset: T.Optional[T.List[str]] = None,
        dtype: T.Type = torch.float32
    ):
        """
        Args:
            pth (Optional[Path], optional): Path to the .tsv or .mgf file containing the mass spectra.
                Default is None, in which case the MassSpecGym dataset is downloaded from HuggingFace Hub.
        """
        self.pth = pth
        self.spec_transform = spec_transform
        self.mol_transform = mol_transform
        self.return_mol_freq = return_mol_freq
        self.return_identifier = return_identifier
        self.identifiers_subset = identifiers_subset
        self.dtype = dtype
        self.load_data()
        self.compute_mol_freq()

    def load_data(self):

        if self.pth is None:
            self.pth = utils.hugging_face_download("MassSpecGym.tsv")

        if isinstance(self.pth, str):
            self.pth = Path(self.pth)

        if self.pth.suffix == ".tsv":
            self.metadata = pd.read_csv(self.pth, sep="\t")
            self.spectra = self.metadata.apply(
                lambda row: matchms.Spectrum(
                    mz=np.array([float(m) for m in row["mzs"].split(",")]),
                    intensities=np.array(
                        [float(i) for i in row["intensities"].split(",")]
                    ),
                    metadata={"precursor_mz": row["precursor_mz"]},
                ),
                axis=1,
            )
            self.metadata = self.metadata.drop(columns=["mzs", "intensities"])
        elif self.pth.suffix == ".mgf":
            self.spectra = pd.Series(list(load_from_mgf(str(self.pth))))
            self.metadata = pd.DataFrame([s.metadata for s in self.spectra])
        else:
            raise ValueError(f"{self.pth.suffix} file format not supported.")
        
        if self.identifiers_subset is not None:
            self.metadata = self.metadata[self.metadata["identifier"].isin(self.identifiers_subset)]
            self.spectra = self.spectra[self.metadata.index].reset_index(drop=True)
            self.metadata = self.metadata.reset_index(drop=True)

    def compute_mol_freq(self):
        if self.return_mol_freq:
            if "inchikey" not in self.metadata.columns:
                self.metadata["inchikey"] = self.metadata["smiles"].apply(utils.smiles_to_inchi_key)
            self.metadata["mol_freq"] = self.metadata.groupby("inchikey")["inchikey"].transform("count")

    def __len__(self) -> int:
        return len(self.spectra)

    def __getitem__(
        self, i: int, transform_spec: bool = True, transform_mol: bool = True
    ) -> dict:
        spec = self.spectra[i]
        metadata = self.metadata.iloc[i]
        mol = metadata["smiles"]

        # Apply all transformations to the spectrum
        item = {}
        if transform_spec and self.spec_transform:
            if isinstance(self.spec_transform, dict):
                for key, transform in self.spec_transform.items():
                    item[key] = transform(spec) if transform is not None else spec
            else:
                item["spec"] = self.spec_transform(spec)
        else:
            item["spec"] = spec

        # Apply all transformations to the molecule
        if transform_mol and self.mol_transform:
            if isinstance(self.mol_transform, dict):
                for key, transform in self.mol_transform.items():
                    item[key] = transform(mol) if transform is not None else mol
            else:
                item["mol"] = self.mol_transform(mol)
        else:
            item["mol"] = mol

        # Add other metadata to the item
        item.update({
            k: metadata[k] for k in ["precursor_mz", "adduct"]
        })

        if self.return_mol_freq:
            item["mol_freq"] = metadata["mol_freq"]

        if self.return_identifier:
            item["identifier"] = metadata["identifier"]

        # TODO: this should be refactored
        for k, v in item.items():
            if not isinstance(v, str):
                item[k] = torch.as_tensor(v, dtype=self.dtype)

        return item

    @staticmethod
    def collate_fn(batch: T.Iterable[dict]) -> dict:
        """
        Custom collate function to handle the outputs of __getitem__.
        """
        return default_collate(batch)


class RetrievalDataset(MassSpecDataset):
    """
    Dataset containing mass spectra and their corresponding molecular structures, with additional
    candidates of molecules for retrieval based on spectral similarity.
    """

    def __init__(
        self,
        mol_label_transform: MolTransform = MolToInChIKey(),
        candidates_pth: T.Optional[T.Union[Path, str]] = None,
        **kwargs,
    ):
        """
        Args:
            mol_label_transform (MolTransform, optional): Transformation to apply to the candidate molecules.
                Defaults to `MolToInChIKey()`.
            candidates_pth (Optional[Union[Path, str]], optional): Path to the .json file containing the candidates for
                retrieval. Defaults to None, in which case the candidates for standard `molecular retrieval` challenge
                are downloaded from HuggingFace Hub. If set to `bonus`, the candidates based on molecular formulas
                for the `bonus chemical formulae challenge` are downloaded instead.
        """
        # note: __init__ calls load_data, these variables are required for load_data to work properly
        self.mol_label_transform = mol_label_transform
        self.candidates_pth = candidates_pth
        super().__init__(**kwargs)

    def load_data(self):

        super().load_data()

        # Download candidates from HuggigFace Hub if not a path to exisiting file is passed
        if self.candidates_pth is None:
            self.candidates_pth = utils.hugging_face_download(
                "molecules/MassSpecGym_retrieval_candidates_mass.json"
            )
        elif self.candidates_pth == 'bonus':
            self.candidates_pth = utils.hugging_face_download(
                "molecules/MassSpecGym_retrieval_candidates_formula.json"
            )
        elif isinstance(self.candidates_pth, str):
            if Path(self.candidates_pth).is_file():
                self.candidates_pth = Path(self.candidates_pth)
            else:
                self.candidates_pth = utils.hugging_face_download(self.candidates_pth)

        # Read candidates_pth from json to dict: SMILES -> respective candidate SMILES
        with open(self.candidates_pth, "r") as file:
            self.candidates = json.load(file)

    def __getitem__(self, i) -> dict:
        item = super().__getitem__(i, transform_mol=False)

        # Save the original SMILES representation of the query molecule (for evaluation)
        item["smiles"] = item["mol"]

        # Get candidates
        if item["mol"] not in self.candidates:
            raise ValueError(f'No candidates for the query molecule {item["mol"]}.')
        item["candidates"] = self.candidates[item["mol"]]

        # Save the original SMILES representations of the canidates (for evaluation)
        item["candidates_smiles"] = item["candidates"]

        # Create neg/pos label mask by matching the query molecule with the candidates
        item_label = self.mol_label_transform(item["mol"])
        item["labels"] = [
            self.mol_label_transform(c) == item_label for c in item["candidates"]
        ]

        if not any(item["labels"]):
            raise ValueError(
                f'Query molecule {item["mol"]} not found in the candidates list.'
            )

        # TODO: it should be refactored this way to a dict in the MassSpecDataset constructor
        #       we don't refactor it now to avoid breaking changes
        if not isinstance(self.mol_transform, dict):
            mol_transform = {"mol": self.mol_transform}
        else:
            mol_transform = self.mol_transform

        # Transform the query and candidate molecules
        for key, transform in mol_transform.items():
            item[key] = transform(item["mol"]) if transform is not None else item["mol"]
            item["candidates_"+key] = [transform(c) if transform is not None else c for c in item["candidates"]]
            if isinstance(item[key], np.ndarray):
                item[key] = torch.as_tensor(item[key], dtype=self.dtype)
                item["candidates_"+key] = torch.as_tensor(np.stack(item["candidates_"+key]), dtype=self.dtype)
        del item["candidates"]

        return item

    @staticmethod
    def _collate_fn_variable_size(batch: T.Iterable[dict], key: str) -> T.Union[torch.Tensor, list]:

        # Flatten the list of lists
        if isinstance(batch[0][key], list):
            collated_item = sum([item[key] for item in batch], start=[])

            # Convert to tensor if it's not a list of strings
            if not isinstance(batch[0][key][0], str):
                if isinstance(batch[0][key][0], Data):
                    # PyG Data object
                    collated_item = Batch.from_data_list(collated_item)
                else:
                    # Standard torch tensor
                    collated_item = torch.as_tensor(collated_item)
        
        # Concatenate the tensors
        elif isinstance(batch[0][key], torch.Tensor):
            collated_item = torch.cat([item[key] for item in batch], dim=0)
        
        # Raise an error if the type is not supported
        else:
            raise ValueError(f"Unsupported type: {type(batch[0][key])}")
        
        return collated_item


    @staticmethod
    def collate_fn(batch: T.Iterable[dict]) -> dict:
        # Standard collate for everything except candidates and their labels (which may have different length per sample)
        collated_batch = {}
        for k in batch[0].keys():
            if k.startswith("candidates") or k == "labels":
                # Handle candidates and labels of variable size per sample
                collated_batch[k] = RetrievalDataset._collate_fn_variable_size(batch, k)
            else:
                # Standard torch or PyG collate for everything else
                if isinstance(batch[0][k], Data):
                    collated_batch[k] = Batch.from_data_list([item[k] for item in batch])
                else:
                    collated_batch[k] = default_collate([item[k] for item in batch])

        # Store the batch pointer reflecting the number of candidates per sample
        collated_batch["batch_ptr"] = torch.as_tensor(
            [len(item["candidates_smiles"]) for item in batch]
        )

        return collated_batch


class SimulationDataset(MassSpecDataset):

    def __init__(
        self,
        spec_transform: SpecTransform,
        mol_transform: MolTransform,
        meta_transform: MetaTransform,
        meta_keys: T.List[str],
        pth: T.Optional[Path] = None,
        return_mol_freq: bool = True,
        return_identifier: bool = True,
        dtype: T.Type = torch.float32
    ): 
        
        # note: __init__ calls load_data, these variables are required for load_data to work properly
        self.meta_transform = meta_transform
        self.meta_keys = meta_keys
        super().__init__(
            spec_transform=spec_transform,
            mol_transform=mol_transform,
            pth=pth,
            return_mol_freq=return_mol_freq,
            return_identifier=return_identifier,
            dtype=dtype
        )

    def load_data(self):

        # set self.pth correctly
        # download if necessary
        if self.pth is None:
            self.pth = utils.hugging_face_download(
                "MassSpecGym.tsv"
            )
        else: 
            assert isinstance(self.pth, str)
            if not os.path.isfile(self.pth):
                self.pth = utils.hugging_face_download(self.pth)

        # will never download here
        super().load_data()

        # remove any spectra not included in the simulation challenge
        sim_mask = self.metadata["simulation_challenge"]
        sim_metadata = self.metadata[sim_mask].copy(deep=True)
        # verify all datapoints are not missing CE information and are [M+H]+
        assert (sim_metadata["adduct"]=="[M+H]+").all()
        assert (~sim_metadata["collision_energy"].isna()).all()
        # mz checks
        assert (sim_metadata["precursor_mz"] <= self.spec_transform.mz_to).all()
        # do the filtering
        self.spectra = self.spectra[sim_mask]
        self.metadata = sim_metadata.reset_index(drop=True) 

    def _get_spec_feats(self, i):

        spectrum = self.spectra.iloc[i]
        spec_feats = self.spec_transform(spectrum)
        return spec_feats

    def _get_mol_feats(self, i):

        metadata = self.metadata.iloc[i]
        mol_feats = self.mol_transform(metadata["smiles"])
        return mol_feats

    def _get_meta_feats(self, i):

        metadata = self.metadata.iloc[i]
        meta_feats = self.meta_transform({k: metadata[k] for k in self.meta_keys})
        return meta_feats

    def _get_other_feats(self, i):

        metadata = self.metadata.iloc[i]
        other_feats = {}
        other_feats["smiles"] = metadata["smiles"]
        if self.return_mol_freq:
            other_feats["mol_freq"] = torch.tensor(metadata["mol_freq"])
        if self.return_identifier:
            other_feats["identifier"] = metadata["identifier"]
        return other_feats

    def __getitem__(self, i) -> dict:
        item = {}
        item.update(self._get_spec_feats(i))
        item.update(self._get_mol_feats(i))
        item.update(self._get_meta_feats(i))
        item.update(self._get_other_feats(i))
        return item
    
    def get_collate_data(self, batch_data: dict) -> dict:

        collate_data = {}
        # handle spectrum
        collate_data.update(self.spec_transform.collate_fn(batch_data))
        # handle molecule
        collate_data.update(self.mol_transform.collate_fn(batch_data))
        # handle metadata
        collate_data.update(self.meta_transform.collate_fn(batch_data))
        # handle other stuff
        if "smiles" in batch_data:
            collate_data["smiles"] = batch_data["smiles"].copy()
        if "mol_freq" in batch_data:
            collate_data["mol_freq"] = torch.stack(batch_data["mol_freq"],dim=0)
        if "identifier" in batch_data:
            collate_data["identifier"] = batch_data["identifier"].copy()
        return collate_data

    def collate_fn(self, data_list: T.List[dict]) -> dict:

        keys = list(data_list[0].keys())
        collate_data = {}
        batch_data = {key: [] for key in keys}
        for data in data_list:
            for key in keys:
                batch_data[key].append(data[key])
        collate_data = self.get_collate_data(batch_data)
        return collate_data

        
class RetrievalSimulationDataset(SimulationDataset):

    def __init__(
        self,
        mol_label_transform: MolTransform = MolToInChIKey(),
        candidates_pth: T.Optional[T.Union[Path, str]] = None,

        **kwargs,
    ):
        # note: __init__ calls load_data, these variables are required for load_data to work properly
        self.mol_label_transform = mol_label_transform
        self.candidates_pth = candidates_pth
        super().__init__(**kwargs)

    def load_data(self):

        super().load_data()

        # Download candidates from HuggigFace Hub
        if self.candidates_pth is None:
            self.candidates_pth = utils.hugging_face_download(
                "molecules/MassSpecGym_retrieval_candidates_mass.json"
            )
        else: 
            assert isinstance(self.candidates_pth, str)
            if not os.path.isfile(self.candidates_pth):
                self.candidates_pth = utils.hugging_face_download(self.candidates_pth)

        # Read candidates_pth from json to dict: SMILES -> respective candidate SMILES
        with open(self.candidates_pth, "r") as file:
            self.candidates = json.load(file)

        # check that everything has candidates
        smileses = self.metadata["smiles"]
        candidates_mask = []
        for smiles in smileses:
            candidates_mask.append(smiles in self.candidates)
        candidates_mask = np.array(candidates_mask)
        assert candidates_mask.all()

    def __getitem__(self, i):

        item = super().__getitem__(i)
        smiles = item["smiles"]
        assert isinstance(smiles, str)

        # Get candidates
        if smiles not in self.candidates:
            raise ValueError(f'No candidates for the query molecule {smiles}.')
        candidates_smiles = self.candidates[smiles]

        # # Save the original SMILES representations of the canidates (for evaluation)
        # item["candidates_smiles"] = candidates_smiles

        # Create neg/pos label mask by matching the query molecule with the candidates
        item_label = self.mol_label_transform(smiles)
        candidates_labels = [
            self.mol_label_transform(c) == item_label for c in candidates_smiles
        ]
        if not any(candidates_labels):
            raise ValueError(
                f'Query molecule {smiles} not found in the candidates list.'
            )
        # item["candidates_labels"] = torch.tensor(candidates_labels)

        candidates_mol_feats, candidates_mask = [], []
        for c in candidates_smiles:
            try:
                candidates_mol_feats.append(self.mol_transform(c))
                candidates_mask.append(True)
            except IndexError as e:
                print(f"> error processing candidate {c} for query {smiles}")
                candidates_mol_feats.append(None)
                candidates_mask.append(False)
        
        candidates_smiles = [candidates_smiles[i] for i in range(len(candidates_smiles)) if candidates_mask[i]]
        candidates_labels = [candidates_labels[i] for i in range(len(candidates_labels)) if candidates_mask[i]]
        candidates_mol_feats = [candidates_mol_feats[i] for i in range(len(candidates_mol_feats)) if candidates_mask[i]]

        # filter based on mask
        item["candidates_smiles"] = candidates_smiles
        item["candidates_labels"] = torch.tensor(candidates_labels)
        item["candidates_mol_feats"] = candidates_mol_feats

        return item

    def collate_fn(self, data_list: T.List[dict]) -> dict:

        keys = list(data_list[0].keys())
        collate_data = {}
        batch_data = {key: [] for key in keys}
        for data in data_list:
            for key in keys:
                batch_data[key].append(data[key])
        collate_data = super().get_collate_data(batch_data)
        # transform candidates mols
        c_collate_data = {}
        c_mol_feats = flatten_lol(batch_data["candidates_mol_feats"])
        c_mol_keys = list(c_mol_feats[0].keys())
        c_mol_batch_data = {key: [] for key in c_mol_keys}
        for c_mol_feats in c_mol_feats:
            for key in c_mol_keys:
                c_mol_batch_data[key].append(c_mol_feats[key])
        c_mol_collate_data = self.mol_transform.collate_fn(c_mol_batch_data)
        # c_meta_feats = batch_data["candidates_meta_feats"]
        # c_meta_keys = list(c_meta_feats[0].keys())
        # c_meta_batch_data = 
        # package it
        prefix = "" # "candidates_"
        for key in c_mol_keys:
            c_collate_data[prefix+key] = c_mol_collate_data[key]
        c_collate_data[prefix+"smiles"] = flatten_lol(batch_data["candidates_smiles"])
        c_collate_data[prefix+"batch_ptr"] = torch.tensor([len(item) for item in batch_data["candidates_smiles"]])
        c_collate_data[prefix+"labels"] = torch.cat(batch_data["candidates_labels"],dim=0)
        # copy relevant keys
        collate_data["candidates_data"] = c_collate_data
        return collate_data


# TODO: Dataset for unlabeled spectra.


import h5py
from tqdm import tqdm
from torch_geometric.data import InMemoryDataset, download_url
from torch_geometric.data import Batch

class MoleculeDataset(InMemoryDataset):
    def __init__(
            self,
            root: str,
            pre_transform: MolTransform,
            pre_filter: T.Callable = lambda x: x is not None,
            url: str = "https://huggingface.co/datasets/roman-bushuiev/MassSpecGym/resolve/main/data/molecules/MassSpecGym_molecules_MCES2_disjoint_with_test_fold_4M.tsv",
            mol_col: str = "smiles",
            verbose: bool = True
        ):
        """
        Args:
            root (str): Root directory where the dataset should be saved.
            pre_transform (MolTransform): Pre-transformation to apply to the molecules.
            pre_filter (Callable, optional): Pre-filter to apply to the pre-transformed molecules. Defaults to lambda x: x is not None.
            url (str, optional): URL to download the .tsv dataset from. Defaults to "https://huggingface.co/datasets/roman-bushuiev/MassSpecGym/resolve/main/data/molecules/MassSpecGym_molecules_MCES2_disjoint_with_test_fold_4M.tsv".
            mol_col (str, optional): Column name of the molecules in the dataset. Defaults to "smiles".
            verbose (bool, optional): Whether to print verbose output. Defaults to True.

        TODO: splits
        """
        self.url = url
        self.mol_col = mol_col
        self.verbose = verbose
        self.file_name = Path(url).name
        super().__init__(root, None, pre_transform, pre_filter)
        self.data, self.slices = torch.load(self.processed_paths[0])

    @property
    def raw_file_names(self):
        return [self.file_name]

    @property
    def processed_file_names(self):
        return [Path(self.file_name).with_suffix(".pt")]

    def download(self):
        download_url(self.url, self.raw_dir)

    def process(self):
        # Read data
        df = pd.read_csv(self.raw_paths[0], sep="\t")
        
        # Process all SMILES strings
        smiles_list = df[self.mol_col].tolist()
        data_list = []
        
        # Process each SMILES string
        for smiles in tqdm(smiles_list, desc="Processing molecules"):
            # Pre-transform
            if self.pre_transform is not None:
                data = self.pre_transform(smiles)
            else:
                data = smiles

            # Pre-filter
            if self.pre_filter is not None:
                if not self.pre_filter(data):
                    continue
                    
            data_list.append(data)

        # Collate all molecules into single batch
        if self.verbose:
            print('Collating...')
        processed_data = Batch.from_data_list(data_list)
        del data_list

        if self.verbose:
            print(f'Total processed molecules: {len(processed_data.x):,}')
            print(f'Saving to {h5_path}...')
        
        h5_path = str(Path(self.processed_paths[0]).with_suffix('.hdf5'))
        with h5py.File(h5_path, 'w') as f:
            # Save all attributes with compression
            for key in processed_data.keys():  # ['x', 'edge_index', 'edge_attr', 'batch', 'ptr']
                data = processed_data[key]
                if torch.is_tensor(data) and data is not None:
                    f.create_dataset(key, data=data.numpy(), compression='gzip', compression_opts=5)

        # Store final result 
        self.save(processed_data, self.processed_paths[0])



import pytest


def simulation_baseline_training():
    import os
    import torch
    from torch.utils.data import DataLoader
    import pytorch_lightning as pl
    import os
    import wandb

    from massspecgym.data.datasets import SimulationDataset, RetrievalSimulationDataset
    from massspecgym.data.transforms import SpecToMzsInts, MolToPyG, StandardMeta, MolToFingerprints
    from massspecgym.models.simulation.fp import FPSimulationMassSpecGymModel
    from massspecgym.models.simulation.gnn import GNNSimulationMassSpecGymModel
    from massspecgym.models.simulation.prec_only import PrecOnlySimulationMassSpecGymModel
    from massspecgym.simulation_utils.run_utils import load_config, get_split_ss

    # default settings for simulation runs
    template_fp = "config/simulation/template.yml"
    # special settings for this demo
    custom_fp = "config/simulation/demo.yml"

    # W&B logging is disabled
    wandb_mode = "disabled"

    config_d = load_config(template_fp,custom_fp)

    pl.seed_everything(config_d["seed"], workers=True)

    # set torch multiprocessing strategy
    torch.multiprocessing.set_sharing_strategy(config_d["mp_sharing_strategy"])

    spec_transform = SpecToMzsInts(
        mz_from=config_d["mz_from"],
        mz_to=config_d["mz_to"],
    )
    if config_d["model_type"] in ["fp", "prec_only"]:
        mol_transform = MolToFingerprints(
            fp_types=config_d["fp_types"]
        )
    elif config_d["model_type"] == "gnn":
        mol_transform = MolToPyG()
    else:
        raise ValueError(f"model_type {config_d['model_type']} not supported")
    meta_transform = StandardMeta(
        adducts=config_d["adducts"],
        instrument_types=config_d["instrument_types"],
        max_collision_energy=config_d["max_collision_energy"]
    )

    # wandb
    wandb.init(
        project=config_d["wandb_project"],
        entity=config_d["wandb_entity"],
        name=config_d["wandb_name"],
        mode=wandb_mode
    )
    logger = pl.loggers.WandbLogger(
        entity=config_d["wandb_entity"],
        project=config_d["wandb_project"],
        name=config_d["wandb_name"],
        mode=wandb_mode,
        tags=[],
        log_model=False,
    )

    # set up df_test_path
    save_df_test = config_d.pop("save_df_test")
    if save_df_test:
        df_test_path = os.path.join(wandb.run.dir, "df_test.pkl")
    else:
        df_test_path = None
    config_d["df_test_path"] = df_test_path

    if config_d["model_type"] == "fp":
        pl_model = FPSimulationMassSpecGymModel(**config_d)
    elif config_d["model_type"] == "prec_only":
        pl_model = PrecOnlySimulationMassSpecGymModel(**config_d)
    elif config_d["model_type"] == "gnn":
        pl_model = GNNSimulationMassSpecGymModel(**config_d)
    else:
        raise ValueError(f"model_type {config_d['model_type']} not supported")
    # print(pl_model)

    ds = SimulationDataset(
        pth=config_d["pth"],
        meta_keys=config_d["meta_keys"],
        spec_transform=spec_transform,
        mol_transform=mol_transform,
        meta_transform=meta_transform
    )

    train_ds, val_ds, test_ds = get_split_ss(
        ds,
        config_d["split_type"],
        subsample_frac=config_d["subsample_frac"]
    )

    dl_config = {
        "num_workers": config_d["num_workers"],
        "batch_size": config_d["batch_size"],
        "drop_last": config_d["drop_last"],
        "pin_memory": config_d["pin_memory"] and config_d["accelerator"] != "cpu",
        "persistent_workers": config_d["persistent_workers"] and config_d["accelerator"] != "cpu",
        "collate_fn": ds.collate_fn
    }

    train_dl = DataLoader(train_ds, shuffle=True, **dl_config)
    val_dl = DataLoader(val_ds, shuffle=False, **dl_config)
    test_dl = DataLoader(test_ds, shuffle=False, **dl_config)

    # NOTE: in this demo, we have disabled the retrieval task
    if config_d["do_retrieval"]:
        # we don't need to create separate datasets, can just overwrite...
        ret_ds = RetrievalSimulationDataset(
            pth=config_d["pth"],
            meta_keys=config_d["meta_keys"],
            spec_transform=spec_transform,
            mol_transform=mol_transform,
            meta_transform=meta_transform,
            candidates_pth=config_d["candidates_pth"]
        )
        ret_dl_config = dl_config.copy()
        ret_dl_config["batch_size"] = config_d["retrieval_batch_size"]
        ret_dl_config["collate_fn"] = ret_ds.collate_fn
        _, _, test_ret_ds = get_split_ss(
            ret_ds,
            config_d["split_type"],
            subsample_frac=config_d["subsample_frac"]
        )
        test_dl = DataLoader(test_ret_ds, shuffle=False, **ret_dl_config)


    # Init trainer
    trainer = pl.Trainer(
        accelerator=config_d["accelerator"], 
        max_epochs=config_d["max_epochs"], 
        logger=logger, 
        log_every_n_steps=config_d["log_every_n_steps"],
        gradient_clip_val=config_d["gradient_clip_val"],
        gradient_clip_algorithm=config_d["gradient_clip_algorithm"]
    )

    # Train
    trainer.fit(
        pl_model, 
        train_dataloaders=train_dl, 
        val_dataloaders=val_dl
    )

    # Test
    trainer.test(
        pl_model,
        dataloaders=test_dl,
        ckpt_path="best"
    )


def test_simulation_baseline_training():
    try:
        simulation_baseline_training()
    except Exception as e:
        pytest.fail(f"simulation_baseline_training() raised an exception: {e}")



import pytest
import numpy as np
import torch
import matchms
from massspecgym.data.transforms import SpecTokenizer, SpecBinner, MolToFormulaVector


def test_spec_tokenizer():
    spec = matchms.Spectrum(
        mz=np.array(
            [
                45.134823,
                45.13699,
                110.096245,
                130.064972,
                136.111862,
                277.16571,
                289.165924,
                307.177856,
                406.223083,
            ]
        ),
        intensities=np.array(
            [0.01082, 0.01064, 0.17184, 0.1397, 0.00874, 1.0, 0.52842, 0.00793, 0.43696]
        ),
        metadata={"precursor_mz": 406.22},
    )

    # Prepend precursor token
    tokenizer = SpecTokenizer(n_peaks=60, prec_mz_intensity=1.1)
    spec_t = tokenizer(spec)
    assert spec_t.shape == (61, 2)

    # Do not prepend precursor token
    tokenizer = SpecTokenizer(n_peaks=60, prec_mz_intensity=None)
    spec_t = tokenizer(spec)
    assert spec_t.shape == (60, 2)
    assert (
        spec_t[: spec.peaks.mz.shape[0], 0] == torch.from_numpy(spec.peaks.mz)
    ).all()
    assert (
        spec_t[: spec.peaks.intensities.shape[0], 1]
        == torch.from_numpy(spec.peaks.intensities)
    ).all()


def test_spec_binner():
    spec = matchms.Spectrum(
        mz=np.array(
            [
                45.134823,
                45.13699,
                110.096245,
                130.064972,
                136.111862,
                277.16571,
                289.165924,
                307.177856,
                406.223083,
            ]
        ),
        intensities=np.array([0.1, 0.1, 0.1, 0.1, 0.1, 1.0, 0.2, 0.1, 0.2]),
        metadata={"precursor_mz": 406.22},
    )
    binner = SpecBinner(max_mz=1000, bin_width=100, to_rel_intensities=False)
    spec_t = binner(spec)
    assert spec_t.shape == (1000 // 100,)
    assert torch.allclose(
        spec_t,
        torch.tensor([0.2, 0.3, 1.2, 0.1, 0.2, 0.0, 0.0, 0.0, 0.0, 0.0]).double(),
    )


def test_resiniferatoxin_formula_vector():
    transform = MolToFormulaVector()
    # Resiniferatoxin (C37H40O9)
    smiles = "C[C@@H]1C[C@]2([C@H]3[C@H]4[C@]1([C@@H]5C=C(C(=O)[C@]5(CC(=C4)COC(=O)Cc6ccc(c(c6)OC)O)O)C)O[C@](O3)(O2)Cc7ccccc7)C(=C)C"
    vector = transform.from_smiles(smiles)
    
    expected_vector = np.zeros(118, dtype=np.float32)
    expected_vector[transform.element_index["C"]] = 37
    expected_vector[transform.element_index["H"]] = 40
    expected_vector[transform.element_index["O"]] = 9
    
    assert np.array_equal(vector, expected_vector), "Ethanol vector does not match expected output"

def test_water_formula_vector():
    transform = MolToFormulaVector()
    # Water (H2O)
    smiles = "O"
    vector = transform.from_smiles(smiles)
    
    expected_vector = np.zeros(118, dtype=np.float32)
    expected_vector[transform.element_index["H"]] = 2
    expected_vector[transform.element_index["O"]] = 1
    
    assert np.array_equal(vector, expected_vector), "Water vector does not match expected output"

# Test case for molecule with an invalid SMILES
def test_invalid_element():
    transform = MolToFormulaVector()
    smiles = "JkO"  # A hypothetical compound
    with pytest.raises(ValueError, match="Invalid SMILES string: JkO"):
        transform.from_smiles(smiles)



import numpy as np
import torch
import matchms
import matchms.filtering as ms_filters
from rdkit.Chem import AllChem as Chem
from typing import Optional
from abc import ABC, abstractmethod
import torch
from torch_geometric.data import Batch

from massspecgym.simulation_utils.feat_utils import MolGraphFeaturizer, get_fingerprints
from massspecgym.simulation_utils.misc_utils import scatter_reduce
import massspecgym.utils as utils
from massspecgym.definitions import CHEM_ELEMS


class SpecTransform(ABC):
    """
    Base class for spectrum transformations. Custom transformatios should inherit from this class.
    The transformation consists of two consecutive steps:
        1. Apply a series of matchms filters to the input spectrum (method `matchms_transforms`).
        2. Convert the matchms spectrum to a torch tensor (method `matchms_to_torch`).
    """

    @abstractmethod
    def matchms_transforms(self, spec: matchms.Spectrum) -> matchms.Spectrum:
        """
        Apply a series of matchms filters to the input spectrum. Abstract method.
        """

    @abstractmethod
    def matchms_to_torch(self, spec: matchms.Spectrum) -> torch.Tensor:
        """
        Convert a matchms spectrum to a torch tensor. Abstract method.
        """

    def __call__(self, spec: matchms.Spectrum) -> torch.Tensor:
        """
        Compose the matchms filters and the torch conversion.
        """
        return self.matchms_to_torch(self.matchms_transforms(spec))


def default_matchms_transforms(
    spec: matchms.Spectrum,
    n_max_peaks: int = 60,
    mz_from: float = 10,
    mz_to: float = 1000,
) -> matchms.Spectrum:
    spec = ms_filters.select_by_mz(spec, mz_from=mz_from, mz_to=mz_to)
    if n_max_peaks is not None:
        spec = ms_filters.reduce_to_number_of_peaks(spec, n_max=n_max_peaks)
    spec = ms_filters.normalize_intensities(spec)
    return spec


class SpecTokenizer(SpecTransform):
    def __init__(
        self,
        n_peaks: Optional[int] = 60,
        prec_mz_intensity: Optional[float] = 1.1,
        matchms_kwargs: Optional[dict] = None
    ) -> None:
        """
        Stack arrays of mz and intensities into a matrix of shape (num_peaks, 2).
        If the number of peaks is less than `n_peaks`, pad the matrix with zeros.
        If `prec_mz_intensity` is provided, prepend the precursor m/z peak with the specified
        intensity value as an additional row at the start of the matrix.

        Args:
            n_peaks (Optional[int], default=60): Maximum number of peaks to keep. If None, 
                keep all peaks.
            prec_mz_intensity (Optional[float], default=1.1): Intensity value to use for the 
                precursor m/z peak. If None, precursor m/z is not added.
            matchms_kwargs (Optional[dict], default=None): Additional keyword arguments to pass 
                to default_matchms_transforms().
        """
        self.n_peaks = n_peaks
        self.prec_mz_intensity = prec_mz_intensity
        self.matchms_kwargs = matchms_kwargs if matchms_kwargs is not None else {}

    def matchms_transforms(self, spec: matchms.Spectrum) -> matchms.Spectrum:
        return default_matchms_transforms(spec, n_max_peaks=self.n_peaks, **self.matchms_kwargs)

    def matchms_to_torch(self, spec: matchms.Spectrum) -> torch.Tensor:
        spec_t = np.vstack([spec.peaks.mz, spec.peaks.intensities]).T
        if self.prec_mz_intensity is not None:
            spec_t = np.vstack([[spec.metadata["precursor_mz"], self.prec_mz_intensity], spec_t])
        if self.n_peaks is not None:
            spec_t = utils.pad_spectrum(
                spec_t,
                self.n_peaks + 1 if self.prec_mz_intensity is not None else self.n_peaks
            )
        return torch.from_numpy(spec_t)


class SpecBinner(SpecTransform):
    def __init__(
        self,
        max_mz: float = 1005,
        bin_width: float = 1,
        to_rel_intensities: bool = True,
    ) -> None:
        self.max_mz = max_mz
        self.bin_width = bin_width
        self.to_rel_intensities = to_rel_intensities
        if not (max_mz / bin_width).is_integer():
            raise ValueError("`max_mz` must be divisible by `bin_width`.")

    def matchms_transforms(self, spec: matchms.Spectrum) -> matchms.Spectrum:
        return default_matchms_transforms(spec, mz_to=self.max_mz, n_max_peaks=None)

    def matchms_to_torch(self, spec: matchms.Spectrum) -> torch.Tensor:
        """
        Bin the spectrum into a fixed number of bins.
        """
        binned_spec = self._bin_mass_spectrum(
            mzs=spec.peaks.mz,
            intensities=spec.peaks.intensities,
            max_mz=self.max_mz,
            bin_width=self.bin_width,
            to_rel_intensities=self.to_rel_intensities,
        )
        return torch.from_numpy(binned_spec)

    def _bin_mass_spectrum(
        self, mzs, intensities, max_mz, bin_width, to_rel_intensities=True
    ):
        # Calculate the number of bins
        num_bins = int(np.ceil(max_mz / bin_width))

        # Calculate the bin indices for each mass
        bin_indices = np.floor(mzs / bin_width).astype(int)

        # Filter out mzs that exceed max_mz
        valid_indices = bin_indices[mzs <= max_mz]
        valid_intensities = intensities[mzs <= max_mz]

        # Clip bin indices to ensure they are within the valid range
        valid_indices = np.clip(valid_indices, 0, num_bins - 1)

        # Initialize an array to store the binned intensities
        binned_intensities = np.zeros(num_bins)

        # Use np.add.at to sum intensities in the appropriate bins
        np.add.at(binned_intensities, valid_indices, valid_intensities)

        # Generate the bin edges for reference
        # bin_edges = np.arange(0, max_mz + bin_width, bin_width)

        # Normalize the intensities to relative intensities
        if to_rel_intensities:
            binned_intensities /= np.max(binned_intensities)

        return binned_intensities  # , bin_edges


class SpecToMzsInts(SpecTransform):
    """
    Similar to SpecTokenizer, but sparse (without padding).
    """

    def __init__(
        self,
        n_peaks: Optional[int] = None,
        mz_from: Optional[float] = 10.,
        mz_to: Optional[float] = 1000.,
        mz_bin_res: Optional[float] = 0.01
    ) -> None:
        self.n_peaks = n_peaks
        self.mz_from = mz_from
        self.mz_to = mz_to
        self.mz_bin_res = mz_bin_res

    def matchms_transforms(self, spec: matchms.Spectrum) -> matchms.Spectrum:

        return default_matchms_transforms(spec, n_max_peaks=self.n_peaks, mz_from=self.mz_from, mz_to=self.mz_to)        

    def matchms_to_torch(self, spec: matchms.Spectrum) -> dict:
        """
        Returns mzs and ints as a dictionary of torch tensors, each of size `n_peaks`.
        """
        mzs = torch.as_tensor(spec.peaks.mz, dtype=torch.float32)
        ints = torch.as_tensor(spec.peaks.intensities, dtype=torch.float32)
        return {"spec_mzs": mzs, "spec_ints": ints}
    
    def collate_fn(self, batch_data_d: dict) -> dict:
        """
        Add batch indices to the data dictionary. This is to support sparse batching (no padding).
        """
        device = batch_data_d["spec_mzs"][0].device
        counts = torch.tensor([spec_mzs.shape[0] for spec_mzs in batch_data_d["spec_mzs"]],device=device,dtype=torch.int64)
        batch_idxs = torch.repeat_interleave(torch.arange(counts.shape[0],device=device),counts,dim=0)
        collate_data_d = {}
        collate_data_d["spec_mzs"] = torch.cat(batch_data_d["spec_mzs"],dim=0)
        collate_data_d["spec_ints"] = torch.cat(batch_data_d["spec_ints"],dim=0)
        collate_data_d["spec_batch_idxs"] = batch_idxs
        return collate_data_d

class MolTransform(ABC):
    @abstractmethod
    def from_smiles(self, mol: str):
        """
        Convert a SMILES string to a tensor-like representation. Abstract method.
        """

    def __call__(self, mol: str):
        return self.from_smiles(mol)


class MolFingerprinter(MolTransform):
    def __init__(self, type: str = "morgan", fp_size: int = 2048, radius: int = 2):
        if type != "morgan":
            raise NotImplementedError(
                "Only Morgan fingerprints are implemented at the moment."
            )
        self.type = type
        self.fp_size = fp_size
        self.radius = radius

    def from_smiles(self, mol: str):
        mol = Chem.MolFromSmiles(mol)
        return utils.morgan_fp(
            mol, fp_size=self.fp_size, radius=self.radius, to_np=True
        )


class MolToInChIKey(MolTransform):
    def __init__(self, twod: bool = True) -> None:
        self.twod = twod

    def from_smiles(self, mol: str) -> str:
        mol = Chem.MolFromSmiles(mol)
        return utils.mol_to_inchi_key(mol, twod=self.twod)


class MolToFormulaVector(MolTransform):
    def __init__(self):
        self.element_index = {element: i for i, element in enumerate(CHEM_ELEMS)}

    def from_smiles(self, smiles: str):
        mol = Chem.MolFromSmiles(smiles)
        if mol is None:
            raise ValueError(f"Invalid SMILES string: {smiles}")

        # Add explicit hydrogens to the molecule
        mol = Chem.AddHs(mol)

        # Initialize a vector of zeros for the 118 elements
        formula_vector = np.zeros(118, dtype=np.int32)

        # Iterate over atoms in the molecule and count occurrences of each element
        for atom in mol.GetAtoms():
            symbol = atom.GetSymbol()
            if symbol in self.element_index:
                index = self.element_index[symbol]
                formula_vector[index] += 1
            else:
                raise ValueError(f"Element '{symbol}' not found in the list of 118 elements.")

        return formula_vector

    @staticmethod
    def num_elements():
        return len(CHEM_ELEMS)


class MolToPyG(MolTransform):

    def __init__(
        self, 
        pyg_node_feats: list[str] = [
            "a_onehot",
            "a_degree",
            "a_hybrid",
            "a_formal",
            "a_radical",
            "a_ring",
            "a_mass",
            "a_chiral"
        ],
        pyg_edge_feats: list[str] = [
            "b_degree",
        ],
        pyg_pe_embed_k: int = 0,
        pyg_bigraph: bool = True):

        self.pyg_node_feats = pyg_node_feats
        self.pyg_edge_feats = pyg_edge_feats
        self.pyg_pe_embed_k = pyg_pe_embed_k
        self.pyg_bigraph = pyg_bigraph

    def from_smiles(self, mol: str):
        
        mol = Chem.MolFromSmiles(mol)
        mg_featurizer = MolGraphFeaturizer(
            self.pyg_node_feats,
            self.pyg_edge_feats,
            self.pyg_pe_embed_k
        )
        mol_pyg = mg_featurizer.get_pyg_graph(mol,bigraph=self.pyg_bigraph)
        return {"mol_pyg": mol_pyg}
    
    def get_input_sizes(self):

        g = self.from_smiles("CCO")["mol_pyg"]
        size_d = {
            "mol_node_feats_size": g.x.shape[1],
            "mol_edge_feats_size": g.edge_attr.shape[1],
        }
        return size_d

    def collate_fn(self, batch_data_d: dict) -> dict:

        collate_data_d = {}
        collate_data_d["mol_pyg"] = Batch.from_data_list(batch_data_d["mol_pyg"])
        return collate_data_d


class MolToFingerprints(MolTransform):

    def __init__(
        self,
        fp_types: list[str] = [
            "morgan",
            "maccs",
            "rdkit"]):

        self.fp_types = sorted(fp_types)

    def from_smiles(self, mol: str) -> dict:
        
        fps = []
        mol = Chem.MolFromSmiles(mol)
        fps = get_fingerprints(mol, self.fp_types)
        fps = torch.as_tensor(fps, dtype=torch.float32)
        return {"fps": fps}
    
    def get_input_sizes(self) -> dict:

        fps = self.from_smiles("CCO")["fps"]
        return {"fps_input_size": fps.shape[0]} 
    
    def collate_fn(self, batch_data_d: dict, prefix="") -> dict:

        collate_data_d = {}
        collate_data_d["fps"] = torch.stack(batch_data_d["fps"],dim=0)
        return collate_data_d


class MetaTransform(ABC):

    @abstractmethod
    def from_meta(self, metadata: dict) -> dict:
        """
        Convert metadata to a dict of numerical representations. Abstract method.
        """

    def __call__(self, metadata: dict):
        return self.from_meta(metadata)


class StandardMeta(MetaTransform):
    
    def __init__(self, adducts: list[str], instrument_types: list[str], max_collision_energy: float):
        self.adduct_to_idx = {adduct: idx for idx, adduct in enumerate(sorted(adducts))}
        self.inst_to_idx = {inst: idx for idx, inst in enumerate(sorted(instrument_types))}
        self.num_adducts = len(self.adduct_to_idx)
        self.num_instrument_types = len(self.inst_to_idx)
        self.num_collision_energies = int(max_collision_energy)
        self.max_collision_energy = max_collision_energy

    def transform_ce(self, ce):

        ce = np.clip(ce, a_min=0, a_max=int(self.max_collision_energy)-1)
        ce_idx = int(np.around(ce, decimals=0))
        return ce_idx

    def from_meta(self, metadata: dict):
        prec_mz = metadata["precursor_mz"]
        adduct_idx = self.adduct_to_idx.get(metadata["adduct"],self.num_adducts)
        inst_idx = self.inst_to_idx.get(metadata["instrument_type"],self.num_instrument_types)
        ce_idx = self.transform_ce(metadata["collision_energy"])
        meta_d = {
            "precursor_mz": torch.tensor(prec_mz,dtype=torch.float32),
            "adduct": torch.tensor(adduct_idx),
            "instrument_type": torch.tensor(inst_idx),
            "collision_energy": torch.tensor(ce_idx)
        }
        return meta_d

    def get_input_sizes(self):

        size_d = {
            "adduct_input_size": self.num_adducts+1,
            "instrument_type_input_size": self.num_instrument_types+1,
            "collision_energy_input_size": int(self.num_collision_energies)
        }
        return size_d

    @property
    def collate_keys(self):

        return ["adduct","instrument_type","collision_energy","precursor_mz"]

    def collate_fn(self, batch_data_d: dict) -> dict:

        collate_data_d = {}
        for key in self.collate_keys:
            collate_data_d[key] = torch.stack(batch_data_d[key],dim=0)
        return collate_data_d



import typing as T
from abc import ABC

import torch
import pandas as pd
from rdkit import Chem
from rdkit.DataStructs import TanimotoSimilarity
from torchmetrics.aggregation import MeanMetric

from massspecgym.models.base import MassSpecGymModel, Stage
from massspecgym.utils import morgan_fp, mol_to_inchi_key, MyopicMCES


class DeNovoMassSpecGymModel(MassSpecGymModel, ABC):

    def __init__(
        self,
        top_ks: T.Iterable[int] = (1, 10),
        myopic_mces_kwargs: T.Optional[T.Mapping] = None,
        *args,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        
        self.top_ks = top_ks
        self.myopic_mces = MyopicMCES(**(myopic_mces_kwargs or {}))
        self.mol_pred_kind: T.Literal["smiles", "rdkit"] = "smiles"
        # caches of already computed results to avoid expensive re-computations
        self.mces_cache = dict()
        self.mol_2_morgan_fp = dict()

    def on_batch_end(
        self,
        outputs: T.Any,
        batch: dict,
        batch_idx: int,
        stage: Stage
    ) -> None:
        self.log(
            f"{stage.to_pref()}loss",
            outputs['loss'],
            batch_size=batch['spec'].size(0),
            sync_dist=True,
            prog_bar=True,
        )

        if stage in self.log_only_loss_at_stages:
            return
        
        metric_vals = self.evaluate_de_novo_step(
            outputs["mols_pred"],  # (bs, k) list of generated rdkit molecules or SMILES strings
            batch["mol"],  # (bs) list of ground truth SMILES strings
            stage=stage
        )

        if stage == Stage.TEST and self.df_test_path is not None:
            self._update_df_test(metric_vals)

    def evaluate_de_novo_step(
        self,
        mols_pred: list[list[T.Optional[Chem.Mol | str]]],
        mol_true: list[str],
        stage: Stage,
    ) -> dict[str, torch.Tensor]:
        """
        # TODO: refactor to compute only for max(k) and then use the result to obtain the rest by
        subsetting.

        Main evaluation method for the models for de novo molecule generation from mass spectra.

        Args:
            mols_pred (list[list[Mol | str]]): (bs, k) list of generated rdkit molecules or SMILES
                strings with possible Nones if no molecule was generated
            mol_true (list[str]): (bs) list of ground-truth SMILES strings
        """
        # Initialize return dictionary to store metric values per sample
        metric_vals = {}

        # Get SMILES and RDKit molecule objects for all predictions
        if self.mol_pred_kind == "smiles":
            smiles_pred_valid, mols_pred_valid = [], []
            for mols_pred_sample in mols_pred:
                smiles_pred_valid_sample, mols_pred_valid_sample = [], []
                for s in mols_pred_sample:
                    m = Chem.MolFromSmiles(s) if s is not None else None
                    # If SMILES cannot be converted to RDKit molecule, the molecule is set to None
                    smiles_pred_valid_sample.append(s if m is not None else None)
                    mols_pred_valid_sample.append(m)
                smiles_pred_valid.append(smiles_pred_valid_sample)
                mols_pred_valid.append(mols_pred_valid_sample)
            smiles_pred, mols_pred = smiles_pred_valid, mols_pred_valid
        elif self.mol_pred_kind == "rdkit":
            smiles_pred = [
                [Chem.MolToSmiles(m) if m is not None else None for m in ms]
                for ms in mols_pred
            ]
        else:
            raise ValueError(f"Invalid mol_pred_kind: {self.mol_pred_kind}")

        # Auxiliary metric: number of valid molecules
        self._update_metric(
            stage.to_pref() + f"num_valid_mols",
            MeanMetric,
            ([sum([m is not None for m in ms]) for ms in mols_pred],),
            batch_size=len(mols_pred),
        )

        # Get RDKit molecule objects for ground truth
        smile_true = mol_true
        mol_true = [Chem.MolFromSmiles(sm) for sm in mol_true]

        def _get_morgan_fp_with_cache(mol):
            """
            A helper function to retrieve either cached Morgan Fingerprint value, or to compute and cache it
            @param mol: RDKit molecule object
            @return:
            """
            if mol not in self.mol_2_morgan_fp:
                self.mol_2_morgan_fp[mol] = morgan_fp(mol, to_np=False)
            return self.mol_2_morgan_fp[mol]

        # Evaluate top-k metrics
        for top_k in self.top_ks:
            # Get top-k predicted molecules for each ground-truth sample
            smiles_pred_top_k = [smiles_pred_sample[:top_k] for smiles_pred_sample in smiles_pred]
            mols_pred_top_k = [mols_pred_sample[:top_k] for mols_pred_sample in mols_pred]

            # 1. Evaluate minimum common edge subgraph:  
            # Calculate MCES distance between top-k predicted molecules and ground truth and
            # report the minimum distance. The minimum distances for each sample in the batch are
            # averaged across the epoch.
            if stage not in self.no_mces_metrics_at_stages:
                min_mces_dists = []
                mces_thld = 100
                # Iterate over batch
                for preds, true in zip(smiles_pred_top_k, smile_true):
                    # Iterate over top-k predicted molecule samples
                    dists = []
                    for pred in preds:
                        if pred is None:
                            dists.append(mces_thld)
                        else:
                            if (true, pred) not in self.mces_cache:
                                mce_val = self.myopic_mces(true, pred)
                                self.mces_cache[(true, pred)] = mce_val
                            dists.append(self.mces_cache[(true, pred)])
                    min_mces_dists.append(min(min(dists), mces_thld))
                min_mces_dists = torch.tensor(min_mces_dists, device=self.device)

                # Log
                metric_name = stage.to_pref() + f"top_{top_k}_mces_dist"
                self._update_metric(
                    metric_name,
                    MeanMetric,
                    (min_mces_dists,),
                    batch_size=len(min_mces_dists),
                    bootstrap=stage == Stage.TEST
                )
                metric_vals[metric_name] = min_mces_dists

            # 2. Evaluate Tanimoto similarity:
            # Calculate Tanimoto similarity between top-k predicted molecules and ground truth and
            # report the maximum similarity. The maximum similarities for each sample in the batch
            # are averaged across the epoch.
            fps_pred_top_k = [
                [_get_morgan_fp_with_cache(m) if m is not None else None for m in ms]
                for ms in mols_pred_top_k
            ]
            fp_true = [_get_morgan_fp_with_cache(m) for m in mol_true]

            max_tanimoto_sims = []
            # Iterate over batch
            for preds, true in zip(fps_pred_top_k, fp_true):
                # Iterate over top-k predicted molecule samples
                sims = [
                    TanimotoSimilarity(true, pred)
                    if pred is not None else 0
                    for pred in preds
                ]
                max_tanimoto_sims.append(max(sims))
            max_tanimoto_sims = torch.tensor(max_tanimoto_sims, device=self.device)

            # Log
            metric_name = stage.to_pref() + f"top_{top_k}_max_tanimoto_sim"
            self._update_metric(
                metric_name,
                MeanMetric,
                (max_tanimoto_sims,),
                batch_size=len(max_tanimoto_sims),
                bootstrap=stage == Stage.TEST
            )
            metric_vals[metric_name] = max_tanimoto_sims

            # 3. Evaluate exact match (accuracy):
            # Calculate if the ground truth molecule is in the top-k predicted molecules and report
            # the average across the epoch.
            in_top_k = [
                mol_to_inchi_key(true) in [
                    mol_to_inchi_key(pred)
                    if pred is not None else None
                    for pred in preds
                ]
                for true, preds in zip(mol_true, mols_pred_top_k)
            ]
            in_top_k = torch.tensor(in_top_k, device=self.device)

            # Log
            metric_name = stage.to_pref() + f"top_{top_k}_accuracy"
            self._update_metric(
                metric_name,
                MeanMetric,
                (in_top_k,),
                batch_size=len(in_top_k),
                bootstrap=stage == Stage.TEST
            )
            metric_vals[metric_name] = in_top_k

        return metric_vals


    def test_step(
        self,
        batch: dict,
        batch_idx: torch.Tensor
    ) -> tuple[torch.Tensor, torch.Tensor]:
        outputs = super().test_step(batch, batch_idx)
        
        # Get generated (i.e., predicted) SMILES
        if self.df_test_path is not None:
            self._update_df_test({
                'identifier': batch['identifier'],
                'mols_pred': outputs['mols_pred']
            })

        return outputs

    def on_test_epoch_end(self):
        # Save test data frame to disk
        if self.df_test_path is not None:
            df_test = pd.DataFrame(self.df_test)
            self.df_test_path.parent.mkdir(parents=True, exist_ok=True)
            df_test.to_pickle(self.df_test_path)



import torch

from massspecgym.data.transforms import MolToFingerprints, StandardMeta
from massspecgym.models.simulation.base import SimulationMassSpecGymModel
from massspecgym.simulation_utils.model_utils import FPModel


class FPSimulationMassSpecGymModel(SimulationMassSpecGymModel):

    def _get_input_sizes(self):

        mol_transform = MolToFingerprints(
            fp_types=self.hparams.fp_types
        )
        meta_transform = StandardMeta(
            adducts=self.hparams.adducts,
            instrument_types=self.hparams.instrument_types,
            max_collision_energy=self.hparams.max_collision_energy)
        input_d = {
            **mol_transform.get_input_sizes(),
            **meta_transform.get_input_sizes()
        }
        return input_d

    def _setup_model(self):

        input_d = self._get_input_sizes()
        self.model = FPModel(
            # input
            fps_input_size=input_d["fps_input_size"],
            metadata_insert_location=self.hparams.metadata_insert_location,
            collision_energy_input_size=input_d["collision_energy_input_size"],
            collision_energy_insert_size=self.hparams.collision_energy_insert_size,
            adduct_input_size=input_d["adduct_input_size"],
            adduct_insert_size=self.hparams.adduct_insert_size,
            instrument_type_input_size=input_d["instrument_type_input_size"],
            instrument_type_insert_size=self.hparams.instrument_type_insert_size,
            # output
            mz_max=self.hparams.mz_max, ###
            mz_bin_res=self.hparams.mz_bin_res,
            # model
            mlp_hidden_size=self.hparams.mlp_hidden_size,
            mlp_dropout=self.hparams.mlp_dropout,
            mlp_num_layers=self.hparams.mlp_num_layers,
            mlp_use_residuals=self.hparams.mlp_use_residuals,
            ff_prec_mz_offset=self.hparams.ff_prec_mz_offset,
            ff_bidirectional=self.hparams.ff_bidirectional,
            ff_output_map_size=self.hparams.ff_output_map_size
        )



import torch

from massspecgym.models.base import Stage
from massspecgym.models.retrieval.base import RetrievalMassSpecGymModel


class RandomRetrieval(RetrievalMassSpecGymModel):

    def step(
        self, batch: dict, stage: Stage = Stage.NONE
    ) -> tuple[torch.Tensor, torch.Tensor]:
        # Generate random retrieval scores
        scores = torch.rand(batch["candidates_mol"].shape[0]).to(self.device)

        # Random baseline, so we return a dummy loss
        loss = torch.tensor(0.0, requires_grad=True)

        return dict(loss=loss, scores=scores)

    def configure_optimizers(self):
        # No optimizer needed for a random baseline
        return None



import typing as T
import pandas as pd
import numpy as np
import pytorch_lightning as pl
import massspecgym.utils as utils
from pathlib import Path
from typing import Optional
from torch.utils.data.dataset import Subset
from torch.utils.data.dataloader import DataLoader
from massspecgym.data.datasets import MassSpecDataset


class MassSpecDataModule(pl.LightningDataModule):
    """
    Data module containing a mass spectrometry dataset. This class is responsible for loading, splitting, and wrapping
    the dataset into data loaders according to pre-defined train, validation, test folds.
    """

    def __init__(
        self,
        dataset: MassSpecDataset,
        batch_size: int,
        num_workers: int = 0,
        persistent_workers: bool = True,
        split_pth: Optional[Path] = None,
        **kwargs
    ):
        """
        Args:
            split_pth (Optional[Path], optional): Path to a .tsv file with columns "identifier" and "fold",
                corresponding to dataset item IDs, and "fold", containg "train", "val", "test"
                values. Default is None, in which case the split from the `dataset` is used.
        """
        super().__init__(**kwargs)
        self.dataset = dataset
        self.split_pth = split_pth
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.persistent_workers = persistent_workers if num_workers > 0 else False

    def prepare_data(self):
        """Pre-processing to be executed only on a single main device when using distributed training."""
        pass

    def setup(self, stage=None):
        """Pre-processing to be executed on every device when using distributed training."""

        # Prepare split
        if self.split_pth is None:
            self.split = self.dataset.metadata[["identifier", "fold"]]
        else:
            # NOTE: custom split is not tested
            self.split = pd.read_csv(self.split_pth, sep="\t")
            if set(self.split.columns) != {"identifier", "fold"}:
                raise ValueError('Split file must contain "id" and "fold" columns.')
            self.split["identifier"] = self.split["identifier"].astype(str)
            if set(self.dataset.metadata["identifier"]) != set(self.split["identifier"]):
                raise ValueError(
                    "Dataset item IDs must match the IDs in the split file."
                )

        self.split = self.split.set_index("identifier")["fold"]
        if not set(self.split) <= {"train", "val", "test"}:
            raise ValueError(
                '"Folds" column must contain only "train", "val", or "test" values.'
            )

        # Split dataset
        split_mask = self.split.loc[self.dataset.metadata["identifier"]].values
        if stage == "fit" or stage is None:
            self.train_dataset = Subset(
                self.dataset, np.where(split_mask == "train")[0]
            )
            self.val_dataset = Subset(self.dataset, np.where(split_mask == "val")[0])
        if stage == "test":
            self.test_dataset = Subset(self.dataset, np.where(split_mask == "test")[0])

    def train_dataloader(self):
        return DataLoader(
            self.train_dataset,
            batch_size=self.batch_size,
            shuffle=True,
            num_workers=self.num_workers,
            persistent_workers=self.persistent_workers,
            drop_last=False,
            collate_fn=self.dataset.collate_fn,
        )

    def val_dataloader(self):
        return DataLoader(
            self.val_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            persistent_workers=self.persistent_workers,
            drop_last=False,
            collate_fn=self.dataset.collate_fn,
        )

    def test_dataloader(self):
        return DataLoader(
            self.test_dataset,
            batch_size=self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            persistent_workers=self.persistent_workers,
            drop_last=False,
            collate_fn=self.dataset.collate_fn,
        )



import numpy as np
import torch
import torch.nn.functional as F
import scipy
import scipy.optimize
from typing import Tuple, Union

from massspecgym.simulation_utils.misc_utils import (
    scatter_logsoftmax, 
    scatter_l1normalize, 
    scatter_l2normalize, 
    scatter_reduce, 
    EPS, 
    th_setdiff1d, 
    safelog, 
    scatter_logsumexp, 
    TOLERANCE_MIN_MZ,
    scatter_logl2normalize,
    LOG_ZERO
)
# rough approximations from NIST20
NCE_MEAN = 60.
NCE_STD = 40.
NCE_MAX = 200.
MZ_MAX = 1500.0
MZ_BIN_RES = 0.01
INTS_THRESH = 0.0
LN_2 = float(np.log(2.))

def get_ints_transform_func(ints_transform):

    _func = None
    if ints_transform == "log10":
        _func = lambda ints: torch.log10(ints+1.)
    elif ints_transform == "log10t3":
        _func = lambda ints: torch.log10(ints/3.+1.)
    elif ints_transform == "loge":
        _func = lambda ints: torch.log(ints+1.)
    elif ints_transform == "sqrt":
        _func = lambda ints: torch.sqrt(ints)
    elif ints_transform == "none":
        _func = lambda ints: ints
    else:
        raise ValueError("Invalid ints_transform: {}".format(ints_transform))
    return _func


def get_ints_untransform_func(ints_transform):

    if ints_transform == "log10":
        max_ints = float(np.log10(1000. + 1.))
        def _untransform_fn(x): return 10**x - 1.
    elif ints_transform == "log10t3":
        max_ints = float(np.log10(1000. / 3. + 1.))
        def _untransform_fn(x): return 3. * (10**x - 1.)
    elif ints_transform == "loge":
        max_ints = float(np.log(1000. + 1.))
        def _untransform_fn(x): return torch.exp(x) - 1.
    elif ints_transform == "sqrt":
        max_ints = float(np.sqrt(1000.))
        def _untransform_fn(x): return x**2
    elif ints_transform == "none":
        max_ints = 1000.
        def _untransform_fn(x): return x
    else:
        raise ValueError("invalid transform")
    def _func(ints, batch_idxs):
        old_max_ints = scatter_reduce(
            ints,
            batch_idxs,
            "amax",
            default=0.,
            include_self=False
        )
        ints = ints / (old_max_ints[batch_idxs] + EPS) * max_ints
        ints = _untransform_fn(ints)
        ints = torch.clamp(ints, min=0.)
        assert not torch.isnan(ints).any()
        return ints
    return _func


def batched_bin_func(
    mzs:torch.Tensor,
    ints:torch.Tensor,
    batch_idxs:torch.Tensor,
    mz_max:float,
    mz_bin_res:float,
    agg:str,
    sparse:bool=False,
    remove_prec_peaks:bool=False,
    prec_mzs:torch.Tensor=None
) -> torch.Tensor:
    """method to get binned spectra for batch

    Args:
        mzs (torch.Tensor): 1d flat tensor of m/zs; multiple m/z lists are concatenated.  
        ints (torch.Tensor): 1d flat tensor of intensities; multiple intensities lists are concatenated.  
        batch_idxs (torch.Tensor): 1d flat tensor of batch indices; each cell, batch_idxs[i] indicates the batch index for mzs[i] and ints[i], should be the same size as mz tensor. 
        mz_max (float): max mz value allowed
        mz_bin_res (float): bin size
        sum_ints (bool): flag for sum intensities within the bin, else take max
        sparse (bool, optional): flag to use sparse  method. Defaults to False.

    Returns:
        _type_: binned spectra
    """

    if mzs.shape[0] == 0:
        import pdb; pdb.set_trace()
    assert torch.max(mzs) < mz_max, (torch.max(mzs),mz_max)
    batch_size = torch.max(batch_idxs)+1
    bins = torch.arange(mz_bin_res,mz_max+mz_bin_res,step=mz_bin_res,device=mzs.device,dtype=mzs.dtype)
    num_bins = bins.shape[0]
    bin_idxs = torch.searchsorted(bins,mzs,side="right")
    bin_offsets = (torch.arange(batch_size,device=mzs.device)*num_bins)[batch_idxs]
    bin_idxs = bin_idxs + bin_offsets
    if remove_prec_peaks:
        assert prec_mzs is not None
        assert torch.max(prec_mzs) <  mz_max, (torch.max(prec_mzs),mz_max)
        prec_mz_bin_idxs = torch.searchsorted(bins,prec_mzs,side="right")
        prec_mz_bin_offsets = torch.arange(batch_size,device=mzs.device)*num_bins
        prec_mz_bin_idxs = prec_mz_bin_idxs + prec_mz_bin_offsets
        prec_ints_mask = torch.isin(bin_idxs,prec_mz_bin_idxs)
        ints = ints*(1-prec_ints_mask.float())
    if sparse:
        un_bin_idxs, un_bin_idxs_rev = torch.unique(bin_idxs,return_inverse=True)
        new_bin_idxs = torch.arange(un_bin_idxs.shape[0],device=un_bin_idxs.device)
        if agg in ["sum","amax"]:
            un_bin_ints = scatter_reduce(
                src=ints,
                index=new_bin_idxs[un_bin_idxs_rev],
                reduce=agg,
                dim_size=new_bin_idxs.shape[0]
            )
        else:
            assert agg == "lse", agg
            un_bin_ints = scatter_logsumexp(
                ints,
                new_bin_idxs[un_bin_idxs_rev],
                dim_size=new_bin_idxs.shape[0]
            )
        un_bin_batch_idxs = un_bin_idxs // num_bins
        return un_bin_idxs, un_bin_ints, un_bin_batch_idxs
    else:
        if agg in ["sum","amax"]:
            bin_spec = scatter_reduce(
                src=ints,
                index=bin_idxs,
                reduce=agg,
                dim_size=num_bins*batch_size
            )
        else:
            assert agg == "lse", agg
            bin_spec = scatter_logsumexp(
                ints,
                bin_idxs,
                dim_size=num_bins*batch_size
            )
        bin_spec = bin_spec.reshape(batch_size,num_bins)
        if agg in ["sum","amax"] and torch.any(torch.all(bin_spec==0.,dim=1)):
            print("> warning: bin_spec is all zeros!")
            mask = torch.zeros_like(bin_spec,dtype=torch.bool)
            mask[:,0] = 1.
            mask = mask*torch.all(bin_spec==0.,dim=1,keepdim=True)
            bin_spec = bin_spec + mask.float()
        return bin_spec

def sparse_cosine_distance(
    true_mzs: torch.Tensor, 
    true_logprobs: torch.Tensor,
    true_batch_idxs: torch.Tensor,
    pred_mzs: torch.Tensor,
    pred_logprobs: torch.Tensor,
    pred_batch_idxs: torch.Tensor,
    mz_max: float,
    mz_bin_res: float,
    log_distance: bool=False
) -> torch.Tensor:

    # sparse bin
    true_bin_idxs, true_bin_logprobs, true_bin_batch_idxs = batched_bin_func(
        true_mzs,
        true_logprobs,
        true_batch_idxs,
        mz_max=mz_max,
        mz_bin_res=mz_bin_res,
        agg="lse",
        sparse=True
    )
    pred_bin_idxs, pred_bin_logprobs, pred_bin_batch_idxs = batched_bin_func(
        pred_mzs,
        pred_logprobs,
        pred_batch_idxs,
        mz_max=mz_max,
        mz_bin_res=mz_bin_res,
        agg="lse",
        sparse=True
    )
    # l2 normalize
    true_bin_logprobs = scatter_logl2normalize(
        true_bin_logprobs,
        true_bin_batch_idxs
    )
    pred_bin_logprobs = scatter_logl2normalize(
        pred_bin_logprobs,
        pred_bin_batch_idxs
    )
    # dot product
    pred_mask = torch.isin(pred_bin_idxs, true_bin_idxs)
    true_mask = torch.isin(true_bin_idxs, pred_bin_idxs)
    batch_size = torch.max(true_bin_batch_idxs)+1
    if torch.any(pred_mask):
        both_bin_logprobs = pred_bin_logprobs[pred_mask] + true_bin_logprobs[true_mask]
        assert torch.all(pred_bin_batch_idxs[pred_mask] == true_bin_batch_idxs[true_mask])
        log_cos_sim = scatter_logsumexp(
            both_bin_logprobs,
            pred_bin_batch_idxs[pred_mask],
            dim_size=batch_size
        )
    else:
        # cosine similarities are all zero
        log_cos_sim = torch.full(
            size=(batch_size,),
            fill_value=LOG_ZERO(pred_bin_logprobs.dtype),
            dtype=pred_bin_logprobs.dtype,
            device=pred_bin_logprobs.device
        )
        # involve pred_logprobs to keep gradient
        log_cos_sim = log_cos_sim + 0.*torch.mean(pred_bin_logprobs,dim=0)
    if log_distance:
        cos_dist = torch.log1p(-torch.exp(log_cos_sim))
    else:
        cos_dist = 1.-torch.exp(log_cos_sim)
    return cos_dist


def batched_l1_normalize(ints, batch_idxs):

    ints = scatter_l1normalize(
        ints,
        batch_idxs
    )
    return ints


def js_sim_helper(
    true_bin_idxs,
    true_bin_ints,
    true_bin_batch_idxs,
    pred_bin_idxs,
    pred_bin_ints,
    pred_bin_batch_idxs
):
    
    batch_size = torch.max(true_bin_batch_idxs)+1
    # l1 normalize
    true_bin_ints = scatter_l1normalize(
        true_bin_ints,
        true_bin_batch_idxs
    )
    pred_bin_ints = scatter_l1normalize(
        pred_bin_ints,
        pred_bin_batch_idxs
    )
    # union distribution
    union_bin_idxs, union_bin_idxs_rev = torch.unique(torch.cat([true_bin_idxs,pred_bin_idxs],dim=0),return_inverse=True)
    union_bin_ints = scatter_reduce(
        src=0.5*torch.cat([true_bin_ints,pred_bin_ints],dim=0),
        index=union_bin_idxs_rev,
        reduce="sum",
        dim_size=union_bin_idxs.shape[0]
    )
    # kl1
    kl1_union_bin_ints = union_bin_ints[union_bin_idxs_rev[:true_bin_idxs.shape[0]]]
    kl1 = scatter_reduce(
        true_bin_ints * (safelog(true_bin_ints) - safelog(kl1_union_bin_ints)),
        true_bin_batch_idxs,
        reduce="sum",
        dim_size=batch_size
    )
    # kl2
    kl2_union_bin_ints = union_bin_ints[union_bin_idxs_rev[true_bin_idxs.shape[0]:]]
    kl2 = scatter_reduce(
        pred_bin_ints * (safelog(pred_bin_ints) - safelog(kl2_union_bin_ints)),
        pred_bin_batch_idxs,
        reduce="sum",
        dim_size=batch_size
    )
    # jss
    jsd = 0.5*(kl1+kl2)
    # jss = (LN_2 - jsd) / LN_2
    jss = 1. - (jsd / LN_2)
    return jss

def sparse_jensen_shannon_similarity(
    true_mzs: torch.Tensor, 
    true_logprobs: torch.Tensor,
    true_batch_idxs: torch.Tensor,
    pred_mzs: torch.Tensor,
    pred_logprobs: torch.Tensor,
    pred_batch_idxs: torch.Tensor,
    mz_max: float,
    mz_bin_res: float
) -> torch.Tensor:
    
    # sparse bin
    true_bin_idxs, true_bin_logprobs, true_bin_batch_idxs = batched_bin_func(
        true_mzs,
        true_logprobs,
        true_batch_idxs,
        mz_max=mz_max,
        mz_bin_res=mz_bin_res,
        agg="lse",
        sparse=True
    )
    pred_bin_idxs, pred_bin_logprobs, pred_bin_batch_idxs = batched_bin_func(
        pred_mzs,
        pred_logprobs,
        pred_batch_idxs,
        mz_max=mz_max,
        mz_bin_res=mz_bin_res,
        agg="lse",
        sparse=True
    )

    jss = js_sim_helper(
        true_bin_idxs,
        true_bin_logprobs.exp(),
        true_bin_batch_idxs,
        pred_bin_idxs,
        pred_bin_logprobs.exp(),
        pred_bin_batch_idxs
    )

    return jss


import time
from functools import wraps
from distutils.util import strtobool
from contextlib import contextmanager
import numpy as np
import torch
import os
import joblib
import tqdm
import pandas as pd
import threading
import torch_geometric as pyg
# import dgl
from collections import defaultdict
import sys

TQDM_DISABLE = False
PPM = 1/1000000
EPS = 1e-9
LOG_HALF = float(np.log(0.5))
LOG_TWO = float(np.log(2.0))
LOG_ZERO_FP32 = float(torch.finfo(torch.float32).min)
LOG_ZERO_FP16 = float(torch.finfo(torch.float16).min)
MAX_CROSS_ENTROPY = 1e19
TOLERANCE_MIN_MZ = 200.0

def LOG_ZERO(dtype):
    if dtype == torch.float32:
        return LOG_ZERO_FP32
    elif dtype == torch.float16:
        return LOG_ZERO_FP16
    else:
        raise ValueError(dtype)

def timeit(func):
    # adapted from https://dev.to/kcdchennai/python-decorator-to-measure-execution-time-54hk
    @wraps(func)
    def timeit_wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        end_time = time.perf_counter()
        total_time = end_time - start_time
        print(f'Function {func.__name__} Took {total_time:.4f} seconds')
        return result
    return timeit_wrapper

def booltype(x):
    return bool(strtobool(x))

def none_or_nan(thing):
    if thing is None:
        return True
    elif isinstance(thing,float) and np.isnan(thing):
        return True
    elif pd.isnull(thing):
        return True
    elif isinstance(thing,str) and thing == "":
        return True
    else:
        return False

@contextmanager
def np_temp_seed(seed):
    state = np.random.get_state()
    np.random.seed(seed)
    try:
        yield
    finally:
        np.random.set_state(state)

@contextmanager
def th_temp_seed(seed):
    state = torch.get_rng_state()
    torch.manual_seed(seed)
    try:
        yield
    finally:
        torch.set_rng_state(state)

def flatten_lol(lol):
    return [item for sublist in lol for item in sublist]

def wandb_symlink(run_dir,wandb_symlink_dp,job_id):
    symlink_dst = os.path.join(wandb_symlink_dp,str(job_id))
    symlink_src = os.path.split(os.path.abspath(run_dir))[0]
    if os.path.islink(symlink_dst):
        os.unlink(symlink_dst)
    os.symlink(symlink_src,symlink_dst)

def list_str2float(str_list):
    return [float(str_item) for str_item in str_list]

# https://stackoverflow.com/a/58936697/6937913
@contextmanager
def tqdm_joblib(tqdm_object):
    """Context manager to patch joblib to report into tqdm progress bar given as argument"""
    class TqdmBatchCompletionCallback(joblib.parallel.BatchCompletionCallBack):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)

        def __call__(self, *args, **kwargs):
            tqdm_object.update(n=self.batch_size)
            return super().__call__(*args, **kwargs)

    old_batch_callback = joblib.parallel.BatchCompletionCallBack
    joblib.parallel.BatchCompletionCallBack = TqdmBatchCompletionCallback
    try:
        yield tqdm_object
    finally:
        joblib.parallel.BatchCompletionCallBack = old_batch_callback
        tqdm_object.close()

# some utils for function timeout
# adapted from https://stackoverflow.com/questions/366682/how-to-limit-execution-time-of-a-function-call

class TimeoutException(Exception): pass

# @contextmanager
# def time_limit(seconds):
# 	if seconds is None:
# 		yield
# 	def signal_handler(signum, frame):
# 		raise TimeoutException(f"Timed out! ({seconds} seconds)")
# 	signal.signal(signal.SIGALRM, signal_handler)
# 	signal.alarm(seconds)
# 	try:
# 		yield
# 	finally:
# 		signal.alarm(0)

def timeout_func(func, args=None, kwargs=None, timeout=30, default=None):
    """This function will spawn a thread and run the given function
    using the args, kwargs and return the given default value if the
    timeout is exceeded.
    http://stackoverflow.com/questions/492519/timeout-on-a-python-function-call
    """
    class InterruptableThread(threading.Thread):
        def __init__(self):
            threading.Thread.__init__(self)
            self.result = default
            self.exc_info = (None, None, None)
        def run(self):
            try:
                self.result = func(*(args or ()), **(kwargs or {}))
            except Exception as err:
                self.exc_info = sys.exc_info()
        def suicide(self):
            raise TimeoutException(
                "{0} timeout (taking more than {1} sec)".format(func.__name__, timeout)
            )
    it = InterruptableThread()
    it.start()
    it.join(timeout)
    if it.exc_info[0] is not None:
        a, b, c = it.exc_info
        raise Exception(a, b, c)  # communicate that to caller
    if it.is_alive():
        it.suicide()
        raise RuntimeError
    else:
        return it.result

def my_tqdm(*args,**kwargs):
    return tqdm.tqdm(*args,**kwargs,disable=TQDM_DISABLE)

def get_tensor_memory_usage(tensor):
    return tensor.nelement()*tensor.element_size()

def get_tensor_dict_memory_usage(**tensor_dict):
    total_memory = 0
    for k,v in tensor_dict.items():
        if isinstance(v,torch.Tensor):
            total_memory += get_tensor_memory_usage(v)
    return total_memory

def get_pyg_memory_usage(pyg_graph):
    return pyg.profile.get_data_size(pyg_graph)

def scatter_masked_softmax(logits,mask,subset_idxs,mask_logprob=None,log=True):
    
    if mask_logprob is None:
        mask_logprob = LOG_ZERO(logits.dtype)	
    # calculate appropriate mask value
    with torch.no_grad():
        c = scatter_masked_logsumexp(logits,mask,subset_idxs)
        lm = torch.gather(
            input=c,
            index=subset_idxs,
            dim=0
        )
        mask_value = mask_logprob + lm
    # apply mask
    masked_logits = mask*logits + (1-mask)*mask_value
    # normalize
    masked_logits = scatter_logsoftmax(masked_logits,subset_idxs)
    if not log:
        # exponentiate
        return torch.exp(masked_logits)
    else:
        return masked_logits

def scatter_masked_logsumexp(logits,mask,subset_idxs,mask_value=None):

    if mask_value is None:
        mask_value = LOG_ZERO(logits.dtype)
    # apply mask
    masked_logits = mask*logits + (1-mask)*mask_value
    # normalize
    masked_logsumexp = scatter_logsumexp(masked_logits,subset_idxs)
    return masked_logsumexp

def scatter_logsumexp(logits,subset_idxs,eps=EPS,dim_size=None):

    if dim_size is None:
        k = torch.max(subset_idxs)+1
    else:
        assert dim_size >= torch.max(subset_idxs)+1
        k = dim_size
    sm = scatter_reduce(
        src=logits,
        index=subset_idxs,
        reduce="amax",
        dim_size=k,
        default=LOG_ZERO(logits.dtype)
    )
    lm = torch.gather(
        input=sm,
        index=subset_idxs,
        dim=0
    )
    logits = logits - lm
    se = scatter_reduce(
        src=torch.exp(logits),
        index=subset_idxs,
        reduce="sum",
        dim_size=k,
        default=0.
    )
    return sm + torch.log(se + eps)

def scatter_logsoftmax(logits,subset_idxs):
    # calculate normalizing constant
    c = scatter_logsumexp(logits,subset_idxs)
    # apply normalizing constant
    logits = logits - c[subset_idxs]
    return logits

def scatter_softmax(logits,subset_idxs):

    return torch.exp(scatter_logsoftmax(logits,subset_idxs))

def scatter_l1normalize(vals,subset_idxs):
    # calculate normalizing constant
    c = scatter_reduce(
        src=vals,
        index=subset_idxs,
        reduce="sum",
        dim_size=torch.max(subset_idxs)+1
    )
    c = torch.clamp(c,min=EPS)
    # apply normalizing constant
    vals = vals/c[subset_idxs]
    return vals

def scatter_l2normalize(vals,subset_idxs):
    # calculate normalizing constant
    c = scatter_reduce(
        src=vals**2,
        index=subset_idxs,
        reduce="sum",
        dim_size=torch.max(subset_idxs)+1
    )
    c = torch.clamp(torch.sqrt(c),min=EPS)
    # apply normalizing constant
    vals = vals/c[subset_idxs]
    return vals

def scatter_logl2normalize(logits,subset_idxs):
    # calculate normalizing constant
    c = scatter_logsumexp(2*logits,subset_idxs)
    # apply normalizing constant
    logits = logits - 0.5*c[subset_idxs]
    return logits

def scatter_var(src,index,dim_size=None,correction=1,sqrt=False):

    # calculate dim_size
    if dim_size is None:
        dim_size = torch.max(index)+1
    else:
        assert dim_size >= torch.max(index)+1
    # calculate mean
    m = scatter_reduce(
        src=src,
        index=index,
        reduce="mean",
        dim_size=dim_size,
        include_self=False
    )
    # calculate variance
    v_num = scatter_reduce(
        src=(src-m[index])**2,
        index=index,
        reduce="sum",
        dim_size=dim_size
    )
    v_den = scatter_reduce(
        src=torch.ones_like(src),
        index=index,
        reduce="sum",
        dim_size=dim_size
    )
    v = v_num/torch.clamp(v_den-correction,min=EPS)
    if sqrt:
        v = torch.sqrt(v)
    return v

def scatter_argmax(src,index,other_index,dim_size=None,return_max=False):

    # calculate dim_size
    if dim_size is None:
        dim_size = torch.max(index)+1
    else:
        assert dim_size >= torch.max(index)+1
    # calculate max
    mx = scatter_reduce(
        src=src,
        index=index,
        reduce="amax",
        dim_size=dim_size,
        include_self=False
    )
    # calculate mask
    ma = src==mx[index]
    # calculate argmax
    ma_idx = other_index*ma + (-1)*(~ma)
    amx = scatter_reduce(
        src=ma_idx,
        index=index,
        reduce="amax",
        dim_size=dim_size,
        include_self=True,
        default=-1
    )
    if return_max:
        return amx, mx
    else:
        return amx

def scatter_reduce(src,index,reduce,dim=0,dim_size=None,default=0.,include_self=True):

    if reduce == "mean" and include_self:
        print("scatter_reduce: mean reduce with include_self=True is not recommended")
    if dim_size is None:
        dim_size = torch.max(index)+1
    else:
        assert dim_size >= torch.max(index)+1
    result_shape = src.shape[:dim] + (dim_size,) + src.shape[dim+1:]
    results = torch.full(result_shape,default,dtype=src.dtype,device=src.device)
    results.scatter_reduce_(
        dim=dim,
        index=index,
        src=src,
        reduce=reduce,
        include_self=include_self
    )
    return results

def safelog(x,eps=EPS):
    return torch.log(torch.clamp(x,min=eps))

def batchwise_max(xs, batch_idxs):
    """ for debugging """

    batch_size = torch.max(batch_idxs)+1
    maxs = torch.zeros([batch_size],device=xs.device,dtype=xs.dtype)
    for b in range(batch_size):
        maxs[b] = torch.max(xs[batch_idxs==b])
    return maxs

def batchwise_lse(xs, batch_idxs):
    """ for debugging """

    batch_size = torch.max(batch_idxs)+1
    lses = torch.zeros([batch_size],device=xs.device,dtype=xs.dtype)
    for b in range(batch_size):
        lses[b] = torch.logsumexp(xs[batch_idxs==b],0)
    return lses

def to_cpu(data_d, non_blocking=True, detach=False):

    for k in data_d.keys():
        if isinstance(data_d[k],torch.Tensor):
            data = data_d[k]
            if detach:
                data = data.detach()
            data = data.to("cpu",non_blocking=non_blocking)
            data_d[k] = data
    return data_d

def to_device(data_d, device, non_blocking=True):

    for k in data_d.keys():
        v = data_d[k]
        # if isinstance(v, torch.Tensor) or isinstance(v, dgl.DGLGraph) or isinstance(v, pyg.data.Data):
        if isinstance(v, torch.Tensor) or isinstance(v, pyg.data.Data):
            v = v.to(device,non_blocking=non_blocking)
            data_d[k] = v
    return data_d

def deep_update(mapping, *updating_mappings):
    """ 
    adapted from pydantic 
    https://github.com/pydantic/pydantic/blob/fd2991fe6a73819b48c906e3c3274e8e47d0f761/pydantic/utils.py#L200-L208 
    """
    updated_mapping = mapping.copy()
    for updating_mapping in updating_mappings:
        for k, v in updating_mapping.items():
            if k in updated_mapping and isinstance(updated_mapping[k], dict) and isinstance(v, dict):
                updated_mapping[k] = deep_update(updated_mapping[k], v)
            else:
                updated_mapping[k] = v
    return updated_mapping

def print_shapes(input_dict):

    for k,v in input_dict.items():
        if isinstance(v,torch.Tensor) or isinstance(v,np.ndarray):
            print(k,"-",tuple(v.shape),"-",type(v))
        elif isinstance(v,list) or isinstance(v,tuple):
            print(k,"-",len(v),"-",type(v))
        elif isinstance(v,pyg.data.Data):
            print(k,"-",(v.num_nodes,v.num_edges),"-",type(v))
        # elif isinstance(v,dgl.DGLGraph):
        # 	print(k,"-",(v.number_of_nodes(),v.number_of_edges()),"-",type(v))
        else:
            print(k,"-",None,"-",type(v))

def th_setdiff1d(t1, t2):

    t1 = torch.unique(t1)
    t2 = torch.unique(t2)
    return t1[(t1[:, None] != t2).all(dim=1)]

def get_package_version(package):

    version = package.__version__.split("+")[0]
    major, minor, patch = version.split(".")
    return (int(major), int(minor), int(patch))

def check_pyg_compile():

    th_major_version, th_minor_version = get_package_version(th)[:2]
    pyg_major_version, pyg_minor_version = get_package_version(pyg)[:2]
    assert th_major_version >= 2, th_major_version
    assert pyg_major_version >= 2, pyg_major_version
    return th_minor_version >= 1 and pyg_minor_version >= 4

def check_pyg_full_compile():

    th_major_version, th_minor_version = get_package_version(th)[:2]
    pyg_major_version, pyg_minor_version = get_package_version(pyg)[:2]
    assert th_major_version >= 2, th_major_version
    return pyg_major_version >= 2 and pyg_minor_version >= 5

class NestedDefaultDict(defaultdict):
    """ https://stackoverflow.com/questions/19189274/nested-defaultdict-of-defaultdict """
    def __init__(self, *args, **kwargs):
        super(NestedDefaultDict, self).__init__(NestedDefaultDict, *args, **kwargs)

    def __repr__(self):
        return repr(dict(self))
